import threading
import uuid


class GenericExecution(threading.Thread):
    def __init__(self, context, name, resolvedCommand, log=None, label=None, debug=False, sensitive=False):
        super().__init__()
        self.threadId = uuid.uuid4()
        self.name = name
        self.resolvedCommand = resolvedCommand
        self.callback = None
        if ('callback' in resolvedCommand):
            self.callback = resolvedCommand['callback']
        self.result = {}
        self.log = log
        self.context = context
        self.label = label
        self.debug = debug
        self.sensitive = sensitive

    def run(self):
        if (self.log and self.debug):
            self.log.log(">> Executing Async")
            self.log.log('')

        # Execute command
        self.result = self.context.executeCommand(self.resolvedCommand, '(In Thread: ' + str(self.threadId) + ')')

        if (self.label):
            self.context.data.storeVar(self.label, self.result['result'])

    def serialize(self):
        return str(self.threadId)



# Generated from D:/Programming/Projects/Apitax/apitax/grammar/src\Ah210.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .Ah210Parser import Ah210Parser
else:
    from Ah210Parser import Ah210Parser

# This class defines a complete generic visitor for a parse tree produced by Ah210Parser.

class Ah210Visitor(ParseTreeVisitor):

    # Visit a parse tree produced by Ah210Parser#prog.
    def visitProg(self, ctx:Ah210Parser.ProgContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#statements.
    def visitStatements(self, ctx:Ah210Parser.StatementsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#statement.
    def visitStatement(self, ctx:Ah210Parser.StatementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#terminated.
    def visitTerminated(self, ctx:Ah210Parser.TerminatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#non_terminated.
    def visitNon_terminated(self, ctx:Ah210Parser.Non_terminatedContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#executers.
    def visitExecuters(self, ctx:Ah210Parser.ExecutersContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#expr.
    def visitExpr(self, ctx:Ah210Parser.ExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#assignment.
    def visitAssignment(self, ctx:Ah210Parser.AssignmentContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#flow.
    def visitFlow(self, ctx:Ah210Parser.FlowContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#if_statement.
    def visitIf_statement(self, ctx:Ah210Parser.If_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#while_statement.
    def visitWhile_statement(self, ctx:Ah210Parser.While_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#for_statement.
    def visitFor_statement(self, ctx:Ah210Parser.For_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#each_statement.
    def visitEach_statement(self, ctx:Ah210Parser.Each_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#condition.
    def visitCondition(self, ctx:Ah210Parser.ConditionContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#block.
    def visitBlock(self, ctx:Ah210Parser.BlockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#callback.
    def visitCallback(self, ctx:Ah210Parser.CallbackContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#callback_block.
    def visitCallback_block(self, ctx:Ah210Parser.Callback_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#optional_parameters_block.
    def visitOptional_parameters_block(self, ctx:Ah210Parser.Optional_parameters_blockContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#sig_parameter.
    def visitSig_parameter(self, ctx:Ah210Parser.Sig_parameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#call_parameter.
    def visitCall_parameter(self, ctx:Ah210Parser.Call_parameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#optional_parameter.
    def visitOptional_parameter(self, ctx:Ah210Parser.Optional_parameterContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#commandtax.
    def visitCommandtax(self, ctx:Ah210Parser.CommandtaxContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#execute.
    def visitExecute(self, ctx:Ah210Parser.ExecuteContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#async_execute.
    def visitAsync_execute(self, ctx:Ah210Parser.Async_executeContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#await.
    def visitAwait(self, ctx:Ah210Parser.AwaitContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#labels.
    def visitLabels(self, ctx:Ah210Parser.LabelsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#label_comp.
    def visitLabel_comp(self, ctx:Ah210Parser.Label_compContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#params_statement.
    def visitParams_statement(self, ctx:Ah210Parser.Params_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#options_statement.
    def visitOptions_statement(self, ctx:Ah210Parser.Options_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#delete_statement.
    def visitDelete_statement(self, ctx:Ah210Parser.Delete_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#error_statement.
    def visitError_statement(self, ctx:Ah210Parser.Error_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#return_statement.
    def visitReturn_statement(self, ctx:Ah210Parser.Return_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#login_statement.
    def visitLogin_statement(self, ctx:Ah210Parser.Login_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#endpoint_statement.
    def visitEndpoint_statement(self, ctx:Ah210Parser.Endpoint_statementContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#scoping.
    def visitScoping(self, ctx:Ah210Parser.ScopingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#name.
    def visitName(self, ctx:Ah210Parser.NameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#exports.
    def visitExports(self, ctx:Ah210Parser.ExportsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#imports.
    def visitImports(self, ctx:Ah210Parser.ImportsContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#casting.
    def visitCasting(self, ctx:Ah210Parser.CastingContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#auth.
    def visitAuth(self, ctx:Ah210Parser.AuthContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#url.
    def visitUrl(self, ctx:Ah210Parser.UrlContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#log.
    def visitLog(self, ctx:Ah210Parser.LogContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#count.
    def visitCount(self, ctx:Ah210Parser.CountContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#inject.
    def visitInject(self, ctx:Ah210Parser.InjectContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#atom.
    def visitAtom(self, ctx:Ah210Parser.AtomContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#obj_dict.
    def visitObj_dict(self, ctx:Ah210Parser.Obj_dictContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#obj_list.
    def visitObj_list(self, ctx:Ah210Parser.Obj_listContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#string.
    def visitString(self, ctx:Ah210Parser.StringContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#number.
    def visitNumber(self, ctx:Ah210Parser.NumberContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by Ah210Parser#boolean.
    def visitBoolean(self, ctx:Ah210Parser.BooleanContext):
        return self.visitChildren(ctx)



del Ah210Parser


# Apitax
# Originally created by Shawn Clake
# shawn.clake@gmail.com
# May 2018
#
# Contains Scriptax and Commandtax
#
# This application is used to automate API processes using arbitrary API's
# It includes a CLI as well as web interface
# However the web interface will need to be customized per API usage

from apitax.ah.Setup import Setup
from apitax.ah.Startup import Startup


class Apitax:

    # Entry point of the program
    def __init__(self, args: list):
        # General procedure is as follows
        # Sets up logical defaults for parameters
        # Checks config for any overrides to those params
        # Checks cli params for any overrides to those params
        # CLI > Config > Default
        setup = Setup(args).load()
        startup = Startup(setup.usage, setup.username, setup.password, setup.watcher, setup.build, setup.script)
        startup.execute(setup.command)





import base64
import json
from apitax.utilities.Files import getAllFiles
from apitax.utilities.Files import readFile
from apitax.utilities.Files import saveFile
from apitax.utilities.Files import deleteFile
from apitax.utilities.Files import renameFile
from apitax.utilities.Files import getPath
from apitax.ah.State import State
from apitax.config.Config import Config as ConfigConsumer


# Base class for driver plugs
# Defines many customizable properties for interfacing to a new API type
class Driver:

    def __init__(self):
        self.config = State.config
        self.driverConfig = None
        if (self.isConfigurable):
            self.driverConfig = ConfigConsumer.read(sectionName=self.__class__.__name__)

    def isConfigurable(self):
        return False

    def getDefaultUsername(self):
        return self.driverConfig.get('default-username')

    def getDefaultPassword(self):
        return self.driverConfig.get('default-password')

    def getAuthEndpoint(self):
        try:
            return self.driverConfig.get('base-endpoint') + self.driverConfig.get('auth-endpoint')
        except:
            return '`base-endpoint` and/or `auth-endpoint` not specified in driver configuration.'

    def getCatalogEndpoint(self):
        try:
            return self.driverConfig.get('base-endpoint') + self.driverConfig.get('catalog-endpoint')
        except:
            return '`base-endpoint` and/or `catalog-endpoint` not specified in driver configuration.'

    def getScriptsPath(self, append=''):
        return getPath(self.config.path + '/drivers/plugins/scriptax/' + self.__class__.__name__ + '/' + append)

    def getPasswordAuthHeader(self, credentials):
        if (not self.isApiAuthenticated()):
            return {}
        temp = credentials.username + ':' + credentials.password
        return {'Authorization': 'Basic ' + base64.b64encode(temp.encode('utf-8'))}

    def getPasswordAuthData(self, credentials):
        if (not self.isApiAuthenticated()):
            return {}
        return {'username': credentials.username, 'password': credentials.password}

    def getTokenAuthHeader(self, credentials):
        if (not self.isApiAuthenticated()):
            return {}
        return {'Authorization': 'Token token="' + credentials.token + '"'}

    def getContentTypeJSON(self):
        # if(not self.isAuthenticated()):
        #   return {}
        return {'Content-type': 'application/json'}

    def getToken(self, response):
        return None

    # Whether or not authentication can produce a usable token or
    # whether to use the username and password for each further request
    def isTokenable(self):
        return True

    # Whether or not crendetials or put into the post data or the header of the
    # authentication request
    def isCredentialsPosted(self):
        return False

    def isApiAuthenticated(self):
        return True

    def piggyBackOffApiAuth(self):
        return False

    def apitaxAuth(self, authObj):
        authObj = authObj['credentials']
        try:
            if(authObj.password == self.users[authObj.username]['password']):
                return self.users[authObj.username]['role']
        except:
            return None
        return None


    def getCatalog(self, auth):
        return {"endpoints": {"tests": {"label": "Placeholder Test", "value": "https://jsonplaceholder.typicode.com"}},
                "selected": "https://jsonplaceholder.typicode.com"}

    def getScriptsCatalog(self):
        files = getAllFiles(self.getScriptsPath("scripts/**/*.ah"))
        returner = {"scripts": []}
        for file in files:
            returner['scripts'].append(
                {"label": file.split('/')[-1].split('.')[0].title(), "relative-path": file, "path": getPath(file)})
        # print(returner)
        return returner

    def getCommandsCatalog(self):
        from apitax.drivers.DriverCommandsFactory import DriverCommandsFactory
        customCommands = DriverCommandsFactory.make(self.config.get('driver') + 'Commands')
        customCommands.setup(self.config, None, None, {}, False, False)
        return customCommands.getCatalog()

    def readScript(self, path):
        return readFile(path)

    def renameScript(self, pathOriginal, pathNew):
        return renameFile(pathOriginal, pathNew)

    # Save does update and create
    def saveScript(self, path, content):
        return saveFile(path, content)

    def deleteScript(self, path):
        return deleteFile(path)

    def serialize(self):
        return {"authenticated": self.isApiAuthenticated(), "auth-tokens": self.isTokenable(),
                "auth-endpoint": self.getAuthEndpoint()}



# System import
import json

# Library import
from bottle import route, run, template, request, static_file, post, Bottle

# Application import
from apitax.ah.Connector import Connector
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.logs.Log import Log
from apitax.utilities.Files import getPath
from apitax.ah.Options import Options
from apitax.ah.Credentials import Credentials

import traceback


# from command import Command
# from ahRequests.authentication import *


# BottleServer is used to provide a web interface which routes through Connector
# It is also used to serve css, html, and javascript

@route('/')
def index():
    return static_file('dashboard.html', bottleServer.directory)


# Hosts html file which will be invoked from browser.
@route('/node/<staticFile>')
def serve_node(staticFile):
    return static_file(staticFile, bottleServer.directory + 'node/dist/')


# Hosts html file which will be invoked from browser.
@route('/pages/<staticFile>')
def serve_static_file(staticFile):
    return static_file(staticFile, bottleServer.directory)


# host css files which will be invoked implicitly by your html files.
@route('/pages/css/<cssFile>')
def serve_css_files(cssFile):
    filePath = bottleServer.directory + 'css/'
    return static_file(cssFile, filePath)


# host js files which will be invoked implicitly by your html files.
@route('/pages/js/<jsFile>')
def serve_js_files(jsFile):
    filePath = bottleServer.directory + 'js/'
    return static_file(jsFile, filePath)


# Authentication endpoint is used to facilitate simpler authentication
@route('/apitax/auth/', method='POST')
def execute_api_auth():
    connector = Connector(sensitive=True,
                          credentials=Credentials(username=request.json['user'], password=request.json['pass']))

    if (connector.http.isTokenable()):
        return json.dumps({"status": 201, "auth": "presumed success", "token": connector.token,
                           "body": json.loads(connector.auth.getResponseBody())})
    else:
        return json.dumps({"status": 400,
                           "auth": "authentication does not support tokens, please pass username and password with each API request"})


@route('/dummy/<filename:path>')
def serve_dummy(filename):
    return static_file(filename, root='/directory/to/files')


# Command endpoint is used to facilitate simpler requests
@route('/apitax/command/', method='POST')
def execute_api_command():
    try:
        connector = None

        parameters = []

        if (request.json['parameters']):
            parameters = request.json['parameters']

        credentials = Credentials()
        options = Options(debug=request.json['debug'], sensitive=request.json['sensitive'])

        if ('token' in request.json):
            credentials.token = request.json['token']
        else:
            credentials.username = request.json['user']
            credentials.password = request.json['pass']
            options.sensitive = True

        if ('extra' in request.json):
            credentials.extra = request.json['extra']

        connector = Connector(options=options, credentials=credentials, command=request.json['command'],
                              parameters=parameters)

        commandHandler = connector.execute()

        returner = {"status": commandHandler.getRequest().getResponseStatusCode(),
                    "body": json.loads(commandHandler.getRequest().getResponseBody())}

        if (request.json['debug']):
            returner.update({"log": connector.logBuffer})

        return json.dumps(returner)
    except:
        bottleServer.log.error(traceback.format_exc())
        if ('debug' in request.json and request.json['debug']):
            return json.dumps({"status": 500, "body": {"error": {
                "state": "Uncaught exception occured during processing. To get a larger stack trace, visit the logs.",
                "message": traceback.format_exc(3)}}})
        else:
            return json.dumps({"status": 500, "body": {}})


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/status', method='GET')
def execute_system_status():
    configDict = bottleServer.config.serialize(["driver", "log", "log-file", "log-colorize"])
    configDict.update({'debug': bottleServer.options.debug, 'sensitive': bottleServer.options.sensitive})
    return json.dumps(configDict)


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/driver/status', method='GET')
def execute_system_driver_status():
    driver = LoadedDrivers.getDefaultBaseDriver()
    driverDict = {"driver": {"name": str(driver.__class__.__name__)}}
    driverDict['driver'].update(driver.serialize())
    return json.dumps(driverDict)


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/scripts/catalog', method='GET')
def execute_system_scripts_catalog():
    driver = LoadedDrivers.getDefaultBaseDriver()
    return json.dumps(driver.getScriptsCatalog())


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/catalog', method='GET')
def execute_system_catalog():
    driver = LoadedDrivers.getDefaultBaseDriver()
    auth = None
    if (request.json):
        auth = request.json['token']
    return json.dumps(driver.getCatalog(auth))


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/scripts', method='POST')
def execute_system_scripts():
    driver = LoadedDrivers.getDefaultBaseDriver()
    return json.dumps({'status': 200, "contents": driver.readScript(request.json['file-name'])})


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/scripts/save', method='POST')
def execute_system_scripts_save():
    driver = LoadedDrivers.getDefaultBaseDriver()
    driver.saveScript(request.json['file-name'], request.json['file'])
    return json.dumps({'status': 200, 'file-name': getPath(request.json['file-name'])})


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/scripts/rename', method='POST')
def execute_system_scripts_rename():
    driver = LoadedDrivers.getDefaultBaseDriver()
    if (not driver.renameScript(request.json['file-name-original'], request.json['file-name-new'])):
        return json.dumps({'status': 500, 'message': 'Cannot rename to an existing file.'})
    return json.dumps({'status': 200, 'file-name': getPath(request.json['file-name-new'])})


# Command endpoint is used to facilitate simpler requests
@route('/apitax/system/scripts/delete', method='POST')
def execute_system_scripts_delete():
    driver = LoadedDrivers.getDefaultBaseDriver()
    driver.deleteScript(request.json['file-name'])
    return json.dumps({'status': 200})


class bottleServer():
    directory = 'apitax/ah/web/node/'
    log = Log()
    config = None
    options = None

    def start(self, ip, port, config=None, options=Options(), reloader=False):
        # self.app = Bottle()
        bottleServer.config = config
        bottleServer.options = options
        bottleServer.directory = config.path + '/ah/web/node/'
        run(host=ip, port=port, reloader=reloader, debug=options.debug)



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.execute import Execute  # noqa: E501
from apitax.ah.api.models.response import Response  # noqa: E501
from apitax.ah.api.models.user_auth import UserAuth  # noqa: E501
from apitax.ah.api.test import BaseTestCase


class TestUsersController(BaseTestCase):
    """UsersController integration test stubs"""

    def test_command(self):
        """Test case for command

        Execute a Command
        """
        execute = Execute()
        response = self.client.open(
            '/apitax/2/command',
            method='POST',
            data=json.dumps(execute),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_display_dashboard(self):
        """Test case for display_dashboard

        Displays the user dashboard page
        """
        response = self.client.open(
            '/apitax/2/dashboard',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_endpoint_catalog(self):
        """Test case for endpoint_catalog

        Retrieve the endpoint catalog
        """
        catalog = UserAuth()
        response = self.client.open(
            '/apitax/2/system/endpoint/catalog',
            method='GET',
            data=json.dumps(catalog),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_get_script(self):
        """Test case for get_script

        Retrieve the contents of a script
        """
        query_string = [('name', 'name_example')]
        response = self.client.open(
            '/apitax/2/system/script',
            method='GET',
            content_type='application/json',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_script_catalog(self):
        """Test case for script_catalog

        Retrieve the script catalog
        """
        response = self.client.open(
            '/apitax/2/system/script/catalog',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# System imports
import json
from time import time

# Application imports
from apitax.ah.Connector import Connector

from apitax.grammar.grammartest import GrammarTest
from apitax.utilities.Numbers import round2str
from apitax.utilities.Npm import Npm

from apitax.ah.Credentials import Credentials

from apitax.ah.State import State

def serialize(obj):
    return obj.serialize()

class Startup:

    def __init__(self, usage, username, password, watcher, build, script) -> None:
        super().__init__()
        self.command = ""
        self.t0 = time()
        self.log = State.log
        self.config = State.config
        self.options = State.options

        self.usage = usage
        self.username = username
        self.password = password
        self.watcher = watcher
        self.build = build
        self.script = script

    def execute(self, command):
        self.command = command

        if (self.usage == 'cli'):
            # Authentication is incorporated into Connector

            if (self.options.debug):
                self.log.log(">>> Starting Processing")
                self.log.log("")
                self.log.log("")

            self.connector = Connector(options=self.options, command=self.command,
                              credentials=Credentials(username=self.username, password=self.password),
                              json=True)
            self.result = self.connector.execute()

            # print(str(connector.http.getCatalog()))

            if (self.options.debug):
                self.log.log(">>> Finished Processing")
                self.log.log("")
                self.log.log("")

            if (self.options.debug and self.command.split(' ')[0] == 'script'):
                for t in self.result.getRequest().parser.threads:
                    t.join()
                self.log.log(">> Dumping Current DataStore Status:")
                self.log.log("    * I recommend this website for looking at the data: http://json.parser.online.fr/")
                self.log.log("")
                self.log.log("")
                self.log.log(json.dumps(self.result.getRequest().parser.data.getStatus(), default=serialize))
                self.log.log("")
                self.log.log("")
            # print(result.getRequest().data.getData('5.3.role_assignments.0.links.assignment'))

            if (self.options.debug):
                self.log.log(">> Apitax finished processing in " + round2str(time() - self.t0) + "s")
                self.log.log("")
                self. log.log("")

        elif (self.usage == 'web'):
            from apitax.ah.api.Server import startDevServer
            if (self.build):
                self.log.log(">> Building Website Assets:")
                self.log.log("")
                path = State.paths['node']
                self.npm = Npm(path)
                self.npm.install()
                self.npm.build()
                if (self.watcher):
                    self.npm.buildWatch(True)
                self.log.log("")
                self.log.log(">> Done Building Website Assets")
            self.log.log(">> Booting Up WebServer:")
            self.log.log("")
            self.server = startDevServer(self.config.get("ip"), self.config.get("port"))
            #bSrv.start(config.get("ip"), config.get("port"), config=config, options=options, reloader=reloader)

        elif (self.usage == 'grammar-test'):
            GrammarTest(self.script)

        elif (self.usage == 'feature-test'):
            pass

        elif (self.usage == 'build'):
            self.log.log(">> Building:")
            self.log.log("> This can take several minutes")
            self.log.log("")
            path = State.paths['node']
            self.npm = Npm(path)
            self.npm.install()
            self.npm.build()
            self.log.log("")
            self.log.log(">> Building Complete")


        else:
            self.log.log("### Error: Unknown mode")

        self.log.getLoggerDriver().outputLog()

        # print(">> Apitax finished processing in {0:.2f}s".format(t1 - t0))



# System import
import json
import re

# Library import
import requests

# Application import
from apitax.logs.Log import Log
from apitax.ah.Options import Options


def serialize(obj):
    return obj.serialize()


# Request is the 'legs' of the application.
# It is responsible to act as a facade to the 'requests' library
# This provides additional output, logging, and logic capabilities
class Request:
    def __init__(self, url, headers='', postData='', paramData={}, pathData={}, options=Options(),
                 customResponse=False):
        self.url = url
        self.headers = headers
        self.postData = postData
        self.paramData = paramData
        self.pathData = pathData
        self.options = options
        self.request = None
        self.customResponse = customResponse
        self.log = Log()
        self.humanReadable = self.log.getLoggerSettings().get('human-readable')

    def setDebug(self, debug):
        self.options.debug = debug

    def setSensitive(self, sensitive):
        self.options.sensitive = sensitive

    def setHeaders(self, headers):
        self.headers = headers

    def setUrl(self, url):
        self.url = url

    def setPostData(self, postData):
        self.postData = postData

    def setParamData(self, paramData):
        self.paramData = paramData

    def setPathData(self, pathData):
        self.pathData = pathData

    def getRequestUrl(self):
        return self.request.url

    def getResponse(self):
        return self.request

    def getResponseHeaders(self):
        if (self.customResponse):
            return self.request['headers']
        else:
            return self.request.headers

    def getResponseBody(self):
        if (self.customResponse):
            if (isinstance(self.request['text'], dict)):
                return json.dumps(self.request['text'], separators=(',', ':'), indent=None).replace("\n", "")
            return self.request['text'].replace("\n", "")
        else:
            if (isinstance(self.request.text, dict) or isinstance(self.request.text, list)):
                return json.dumps(self.request.text, separators=(',', ':'), indent=None).replace("\n", "")
            return self.request.text.replace("\n", "")

    def getResponseBodyAsStructure(self):
        return json.loads(self.getResponseBody())

    def getResponseStatusCode(self):
        if (self.customResponse):
            return self.request['status_code']
        else:
            return self.request.status_code

    def getCLIResponse(self):
        return json.dumps(self.getResponseBodyAsStructure(), indent=2, separators=(',', ': '))

    def getDebugResponse(self):
        line = ''
        line += 'Status: ' + str(self.getResponseStatusCode()) + '\n'
        line += 'Headers:' + '\n'
        if (self.options.sensitive):
            line += 'Headers are not shown as it contains sensitive data. ie. token' + '\n'
        else:
            line += str(self.getResponseHeaders()) + '\n'
        line += 'Body:' + '\n'
        line += str(self.getResponseBody()) + '\n'
        return line

    def getDebugRequest(self):
        line = ''
        line += 'Endpoint:        ' + self.url + '\n'
        line += 'Formed Endpoint: ' + self.getRequestUrl() + '\n'
        line += 'Headers:' + '\n'
        if (self.options.sensitive):
            line += 'Headers are not shown as it contains sensitive data. ie. password' + '\n'
        else:
            line += str(self.headers) + '\n'
        line += 'Post Data:' + '\n'
        if (self.options.sensitive):
            line += 'Post Data is not shown as it contains sensitive data. ie. password' + '\n'
        elif (not self.postData):
            line += '{}' + '\n'
        else:
            line += str(self.postData) + '\n'
        return line

    def injectPathData(self):
        if (not self.pathData):
            return
        matches = re.findall('{[A-z0-9]{1,}}', self.url)
        for match in matches:
            matchStr = match[1:-1]
            if (matchStr in self.pathData):
                self.url = self.url.replace(match, self.pathData.get(matchStr))
            else:
                self.log.log('Path data did not contain key for `' + matchStr + '`')

    def logRequest(self):
        if (self.options.debug and not self.humanReadable):
            self.log.log(
                '\n<==========' + '\n' + self.getDebugRequest() + '\n' + self.getDebugResponse() + '==========>' + '\n\n')

        elif (self.options.debug):
            self.log.log('\n<==========\n' + self.getCLIResponse() + '\n' + '==========>\n\n')

    def post(self):
        self.injectPathData()
        self.request = requests.post(self.url, data=json.dumps(self.postData), headers=self.headers,
                                     params=self.paramData)
        self.logRequest()

    def get(self):
        self.injectPathData()
        self.request = requests.get(self.url, data=json.dumps(self.postData), headers=self.headers,
                                    params=self.paramData)
        self.logRequest()

    def put(self):
        self.injectPathData()
        self.request = requests.put(self.url, data=json.dumps(self.postData), headers=self.headers,
                                    params=self.paramData)
        self.logRequest()

    def patch(self):
        self.injectPathData()
        self.request = requests.patch(self.url, data=json.dumps(self.postData), headers=self.headers,
                                      params=self.paramData)
        self.logRequest()

    def delete(self):
        self.injectPathData()
        self.request = requests.delete(self.url, data=json.dumps(self.postData), headers=self.headers,
                                       params=self.paramData)
        self.logRequest()



import connexion
import six

from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.response import Response  # noqa: E501
from apitax.ah.api import util

from apitax.ah.State import State
from apitax.ah.LoadedDrivers import LoadedDrivers

from apitax.ah.api.utilities.Roles import hasAccess as roleAccess
from flask import redirect

from flask_jwt_extended import (create_access_token, create_refresh_token, jwt_required, jwt_refresh_token_required,
                                get_jwt_identity, get_raw_jwt, get_jwt_claims)


@jwt_required
def display_admin_dashboard():  # noqa: E501
    """Displays the developer admin page

    Displays the user admin page # noqa: E501


    :rtype: None
    """
    if(not hasAccess()):
        return redirectUnauthorized()

    return 'do some magic!'


@jwt_required
def driver_status(name):  # noqa: E501
    """Retrieve the status of a loaded driver

    Retrieve the status of a loaded driver # noqa: E501

    :param name: Get status of a driver with this name
    :type name: str

    :rtype: Response
    """
    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    body = {"name": str(driver.__class__.__name__)}
    body.update(driver.serialize())
    return Response(status=200, body=body)


@jwt_required
def system_status():  # noqa: E501
    """Retrieve the system status

    Retrieve the system status # noqa: E501


    :rtype: Response
    """
    if(not hasAccess()):
        return redirectUnauthorized()

    body = State.config.serialize(["driver", "log", "log-file", "log-colorize"])
    body.update({'debug': State.options.debug, 'sensitive': State.options.sensitive})
    return Response(status=200, body=body)


def hasAccess():
    if (roleAccess(get_jwt_identity(), 'admin')):
        return True
    return False


def redirectUnauthorized(page="login", code=303):
    return redirect(State.baseUrl + page, code=code)




from apitax.logs.LogDriver import LogDriver

class BufferedLog(LogDriver):
    def __init__ (self):
        self.buffer = []
        super().__init__()
        
    def log(self, line, type='info'):
        self.buffer.append({'line': line, 'type': type})
        	
    def outputLog(self):
        for line in self.buffer:
            self.logger.info(line['line'])
            print(self.injectStdColor(line['line']))
        self.buffer = []


from apitax.ah.State import State
from apitax.ah.commandtax.ApiAuthentication import AuthRequest
from apitax.ah.LoadedDrivers import LoadedDrivers


class ApitaxAuthentication:

    @staticmethod
    def login(credentials):
        driver = LoadedDrivers.getAuthDriver()
        role = None
        if(driver.piggyBackOffApiAuth()):
            #apiDriver = LoadedDrivers.getPrimaryDriver()
            apiDriver = driver

            if('driver' in credentials.extra):
                apiDriver = LoadedDrivers.getBaseDriver(credentials.extra['driver'])
            apiAuth = AuthRequest(credentials, apiDriver, State.options)
            apiAuth.authenticate()

            if(not apiAuth.isAuthenticated()):
                return None

            role = driver.apitaxAuth({"credentials": credentials, "apiAuthRequest": apiAuth})
            if(apiDriver.isTokenable()):
                credentials.token = apiAuth.getToken()
        else:
            role = driver.apitaxAuth({"credentials": credentials})

        if(not role):
            return None
        return {"role": role, "credentials": credentials}











class Credentials:
    def __init__(self, username='', password='', token='', extra={}):
        self.username = username
        self.password = password
        self.token = token
        self.extra = extra

    def getAuthObj(self):
        return {'username': self.username, 'password': self.password, 'token': self.token}

    def serialize(self):
        return str(self.username)




from apitax.logs.LogDriver import LogDriver

class StandardLog(LogDriver):
    def __init__ (self):
        super().__init__()
        
    def log(self, line, type='info'):
        self.logger.info(line)
        print(self.injectStdColor(line))



import connexion
import six

import flask

from apitax.ah.api.models.auth_response import AuthResponse  # noqa: E501
from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.user_auth import UserAuth  # noqa: E501
from apitax.ah.api import util

from apitax.ah.ApitaxAuthentication import ApitaxAuthentication
from apitax.ah.api.utilities.Mappers import mapUserAuthToCredentials

from flask_jwt_extended import (create_access_token, create_refresh_token, jwt_required, jwt_refresh_token_required, get_jwt_identity, get_raw_jwt)

def authenticate(user=None):  # noqa: E501
    """Authenticate

    Authenticate with the API # noqa: E501

    :param user: The user authentication object.
    :type user: dict | bytes

    :rtype: UserAuth
    """
    if connexion.request.is_json:
        user = UserAuth.from_dict(connexion.request.get_json())  # noqa: E501

    credentials = mapUserAuthToCredentials(user)
    auth = ApitaxAuthentication.login(credentials)
    if(not auth):
        return ErrorResponse(status=401, message="Invalid credentials")

    access_token = create_access_token(identity={'username': user.username, 'role': auth['role']})
    refresh_token = create_refresh_token(identity={'username': user.username, 'role': auth['role']})

    return AuthResponse(status=201, message='User ' + user.username + ' was authenticated as ' + auth['role'], access_token=access_token, refresh_token=refresh_token, auth=UserAuth(username=auth['credentials'].username, api_token=auth['credentials'].token))


def display_login():  # noqa: E501
    """Displays the login page

    Displays the login page # noqa: E501


    :rtype: None
    """
    
    return flask.send_from_directory('dashboard/src/pages', 'landing.html')


def get_asset(name):  # noqa: E501
    """Displays the user dashboard page

    Displays the user dashboard page # noqa: E501

    :param name: Get an asset such as a JavaScript file
    :type name: str

    :rtype: None
    """
    return flask.send_from_directory('dashboard/dist', name)


@jwt_refresh_token_required
def refresh_token():  # noqa: E501
    """Refreshes login token using refresh token

    Refreshes login token using refresh token # noqa: E501


    :rtype: UserAuth
    """
    current_user = get_jwt_identity()
    if(not current_user):
        return ErrorResponse(status=401, message="Not logged in")
    access_token = create_access_token(identity=current_user)
    return AuthResponse(status=201, message='Refreshed Access Token', access_token=access_token, auth=UserAuth())



from apitax.ah.commandtax.Commandtax import Commandtax
from apitax.ah.commandtax.ApiAuthentication import *
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.ah.HeaderBuilder import HeaderBuilder
from apitax.config.Config import Config
from apitax.logs.Log import Log
from apitax.ah.Credentials import Credentials
from apitax.ah.Options import Options

from time import time


# The 'heart' of the application
# Connector facilitates the initialization of the connection to an API
# Connector handles setting up the driver, authetnication, headers, and then
# using all of that to execute a command
#
# Additional interfaces to this utility should directly communicate with connector
# and likely nothing else. Connector handles the rest.
class Connector:

    def __init__(self, options=Options(), credentials=None, command='', json=True, parameters={}):

        self.options = options
        self.parameters = parameters

        self.credentials = credentials

        self.command = command
        self.command = self.command.replace('\\"', '"')
        self.command = self.command.replace('\\\'', '\'')

        self.executionTime = None
        self.commandHandler = None
        self.logBuffer = []

        self.config = Config.read()

        if (self.options.driver):
            self.http = LoadedDrivers.getBaseDriver(self.options.driver)
        else:
            self.http = LoadedDrivers.getDefaultBaseDriver()

        self.header = HeaderBuilder()
        if (json):
            self.header.build(self.http.getContentTypeJSON())

        if (self.credentials.token == ''):
            preHeader = self.header.header.copy()
            if (self.http.isTokenable()):
                auth = AuthRequest(self.credentials, self.http, self.options)
                auth.authenticate()
                self.credentials.token = auth.getToken()
                self.header.header = preHeader
                self.header.build(self.http.getTokenAuthHeader(self.credentials))
            else:
                self.header.build(self.http.getPasswordAuthHeader(self.credentials))
        else:
            self.header.build(self.http.getTokenAuthHeader(self.credentials))

    def getCredentials(self):
        return self.credentials

    def execute(self, command=''):
        if (command != ''):
            self.command = command
        t0 = time()
        self.commandHandler = Commandtax(self.header, self.command, self.config, options=self.options,
                                         parameters=self.parameters, auth=self.getCredentials())

        self.executionTime = time() - t0

        log = Log()
        self.logBuffer = log.getLoggerDriver().buffer
        log.getLoggerDriver().outputLog()

        return self.commandHandler



def isBoolean(v):
    return v.lower() in ("true", "false")


def str2bool(v):
    if (v.lower() in ("true")):
        return True
    elif (v.lower() in ("false")):
        return False
    else:
        return None



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from apitax.ah.api.models.create import Create  # noqa: E501
from apitax.ah.api.models.delete import Delete  # noqa: E501
from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.rename import Rename  # noqa: E501
from apitax.ah.api.models.response import Response  # noqa: E501
from apitax.ah.api.models.save import Save  # noqa: E501
from apitax.ah.api.test import BaseTestCase


class TestDevelopersController(BaseTestCase):
    """DevelopersController integration test stubs"""

    def test_create_script(self):
        """Test case for create_script

        Create a new script
        """
        create = Create()
        response = self.client.open(
            '/apitax/2/system/script',
            method='POST',
            data=json.dumps(create),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_delete_script(self):
        """Test case for delete_script

        Delete a script
        """
        delete = Delete()
        response = self.client.open(
            '/apitax/2/system/script',
            method='DELETE',
            data=json.dumps(delete),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_display_developer_dashboard(self):
        """Test case for display_developer_dashboard

        Displays the developer dashboard page
        """
        response = self.client.open(
            '/apitax/2/dashboard/developer',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_rename_script(self):
        """Test case for rename_script

        Rename a script
        """
        rename = Rename()
        response = self.client.open(
            '/apitax/2/system/script',
            method='PATCH',
            data=json.dumps(rename),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_save_script(self):
        """Test case for save_script

        Save a script
        """
        save = Save()
        response = self.client.open(
            '/apitax/2/system/script',
            method='PUT',
            data=json.dumps(save),
            content_type='application/json')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# Import Drivers below here

from apitax.drivers.Drivers import Drivers


# End Driver Imports


# Factory class for creating HttpPlug Drivers
class DriverFactory:

    @staticmethod
    def make(name):
        return Drivers.get(name)



import glob
import os
import sys
from pathlib import Path


def getAllFiles(path):
    return glob.glob(getPath(path), recursive=True)


def readFile(path):
    with open(getPath(path), 'r') as myfile:
        return myfile.read()  # .replace('\n', '').replace('\r', '')


def renameFile(originalFile, newFile):
    if (getPath(newFile, asString=False).exists()):
        return False
    getPath(originalFile, asString=False).rename(Path(newFile))
    return True


def saveFile(path, contents):
    with open(getPath(path), "w+") as text_file:
        print(contents, file=text_file)


def deleteFile(path):
    return getPath(path, asString=False).unlink()


def getPath(path, asString=True):
    pathStr = str(Path(path).resolve()).replace(os.sep, '/')
    if (asString):
        return pathStr
    else:
        return Path(pathStr)


def createDir(path):
    Path(path).mkdir(parents=True, exist_ok=True)


def getRoot(path=''):
    return getPath(os.path.dirname(getPath(sys.modules['__main__'].__file__)) + path)



# System import
import logging
import re
from logging.handlers import RotatingFileHandler
from pathlib import Path

# Logging object to prevent duplicate handler adding
# global loggers
loggers = {}


# Handles printing to CLI as well as printing to log file
class Log:

    def __init__(self, logDriver=None, logFile='logs/apitax.log', doLog=True, logColorize=True, logPrefixes=True, logHumanReadable=False):
        # logging.basicConfig(filename=filepath,level=logging.INFO)
        # log_formatter = logging.Formatter('%(asctime)s %(levelname)s %(funcName)s(%(lineno)d) %(message)s')

        self.prefix = ''

        if (not loggers.get('main')):

            directorypath = Path(logFile)
            # print("/".join(directorypath.parts[:-1]))
            Path("/".join(directorypath.parts[:-1])).mkdir(parents=True, exist_ok=True) 

            log_formatter = logging.Formatter('%(asctime)s %(levelname)s  %(message)s')

            # Each log file can be up to 25MB big and keeps a max of 3 backup files.
            # Thus, the largest size of the log folder is 100MB
            my_handler = RotatingFileHandler(logFile, mode='a', maxBytes=25 * 1024 * 1024,
                                             backupCount=3, encoding=None, delay=0)
            my_handler.setFormatter(log_formatter)
            my_handler.setLevel(logging.INFO)

            app_log = logging.getLogger('root')
            app_log.setLevel(logging.INFO)

            app_log.addHandler(my_handler)

            loggers.update({'main': app_log, 'driver': logDriver, 'settings':{'doLog': doLog, 'colorize':logColorize, 'prefixes':logPrefixes, 'human-readable': logHumanReadable, 'path':directorypath.resolve()}})

            logDriver.setLogger(self.getLogger())
            logDriver.setSettings(self.getLoggerSettings())
                        

    def inject(self, text, prefix=''):
        
        if(prefix == ''):
            prefix = self.prefix
        
        if(prefix == '' or text == ''):
            return text

        text = str(text)

        if (text[:3] == '>>>' or text[:3] == '###'):
            return text[:3] + ' ' + prefix + ' ' + text[3:]
        elif (text[:2] == '>>'):
            return text[:2] + ' ' + prefix + ' ' + text[2:]
        elif (text[:1] == '>' or text.strip()[:1] == '*'):
            return text[:1] + ' ' + prefix + ' ' + text[1:]

        return prefix + ' ' + text

    def isLoggable(self, prefix=''):
        if((prefix != '' or self.prefix != '') and not loggers.get('settings').get('prefixes')):
            return False
        return loggers.get('settings').get('doLog')

    def log(self, text, prefix=''):
        # logging.info(' '+text)
        text = self.inject(text, prefix)
        if(self.isLoggable(prefix)):
            self.getLoggerDriver().log(text)

    def error(self, text, prefix=''):
        # logging.info(' '+text)
        text = self.inject(text, prefix)
        if(self.isLoggable(prefix)):
            self.getLoggerDriver().log('### Error: ' + text)
        
    def getLogger(self):
        return loggers.get('main')
        
    def getLoggerSettings(self):
        return loggers.get('settings')
        
    def getLoggerDriver(self):
        return loggers.get('driver')
        

        



# Base class for driver command classes

from apitax.utilities.Files import getPath
from apitax.utilities.Files import getAllFiles
from apitax.ah.Options import Options


class DriverCommands():

    def setup(self, config, header, auth, parameters, options=Options()):
        self.config = config
        self.header = header
        self.auth = auth
        self.options = options
        self.request = None
        self.parameters = parameters
        if (self.getDriverPath() is None):
            self.basePath = str(config.path) + '/drivers/plugins/commandtax/' + self.getDriverName()
        else:
            self.basePath = self.getDriverPath() + '/' + self.getDriverName()
        self.basePath = getPath(self.basePath)
        return self

    def getDriverName(self):
        return None

    def getDriverPath(self):
        return None

    def getRequest(self):
        return self.request

    def override(self, command: list):
        return False

    def getCatalog(self):
        path = self.basePath
        baseLength = len(path)
        files = getAllFiles(path + "/**/*.ah")
        catalog = []
        for file in files:
            fPath = getPath(file)
            command = fPath[baseLength + 1:-3].replace('/', ' ').replace('\\', ' ')
            catalog.append(command)
        return catalog

    def handle(self, command):
        from apitax.ah.commandtax.commands.Script import Script
        if (self.override(command)):
            return
        self.request = Script(self.config, self.header, self.auth, self.parameters, self.options)
        path = self.basePath
        i = 0
        for element in command:
            element = element.strip()
            if (element == '--driver'):
                i += 2
            elif (element[:2] == '--'):
                i += 1

            if (i > 0):
                i -= 1
            else:
                path += '/' + element
            # print(element+":"+str(i))
        path += '.ah'
        path = getPath(path)
        self.request.handle([path])
        return self.request




from apitax.ah.Credentials import Credentials
from apitax.ah.api.models.user_auth import UserAuth


def mapUserAuthToCredentials(userAuth: UserAuth, credentials=Credentials()):
    if (userAuth.api_token):
        credentials.token = userAuth.api_token
    else:
        credentials.username = userAuth.username
        credentials.password = userAuth.password

    if (userAuth.extra):
        credentials.extra = userAuth.extra

    return credentials




# Generated from D:/Programming/Projects/Apitax/apitax/grammar/src\Ah210.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .Ah210Parser import Ah210Parser
else:
    from Ah210Parser import Ah210Parser

# This class defines a complete listener for a parse tree produced by Ah210Parser.
class Ah210Listener(ParseTreeListener):

    # Enter a parse tree produced by Ah210Parser#prog.
    def enterProg(self, ctx:Ah210Parser.ProgContext):
        pass

    # Exit a parse tree produced by Ah210Parser#prog.
    def exitProg(self, ctx:Ah210Parser.ProgContext):
        pass


    # Enter a parse tree produced by Ah210Parser#statements.
    def enterStatements(self, ctx:Ah210Parser.StatementsContext):
        pass

    # Exit a parse tree produced by Ah210Parser#statements.
    def exitStatements(self, ctx:Ah210Parser.StatementsContext):
        pass


    # Enter a parse tree produced by Ah210Parser#statement.
    def enterStatement(self, ctx:Ah210Parser.StatementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#statement.
    def exitStatement(self, ctx:Ah210Parser.StatementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#terminated.
    def enterTerminated(self, ctx:Ah210Parser.TerminatedContext):
        pass

    # Exit a parse tree produced by Ah210Parser#terminated.
    def exitTerminated(self, ctx:Ah210Parser.TerminatedContext):
        pass


    # Enter a parse tree produced by Ah210Parser#non_terminated.
    def enterNon_terminated(self, ctx:Ah210Parser.Non_terminatedContext):
        pass

    # Exit a parse tree produced by Ah210Parser#non_terminated.
    def exitNon_terminated(self, ctx:Ah210Parser.Non_terminatedContext):
        pass


    # Enter a parse tree produced by Ah210Parser#executers.
    def enterExecuters(self, ctx:Ah210Parser.ExecutersContext):
        pass

    # Exit a parse tree produced by Ah210Parser#executers.
    def exitExecuters(self, ctx:Ah210Parser.ExecutersContext):
        pass


    # Enter a parse tree produced by Ah210Parser#expr.
    def enterExpr(self, ctx:Ah210Parser.ExprContext):
        pass

    # Exit a parse tree produced by Ah210Parser#expr.
    def exitExpr(self, ctx:Ah210Parser.ExprContext):
        pass


    # Enter a parse tree produced by Ah210Parser#assignment.
    def enterAssignment(self, ctx:Ah210Parser.AssignmentContext):
        pass

    # Exit a parse tree produced by Ah210Parser#assignment.
    def exitAssignment(self, ctx:Ah210Parser.AssignmentContext):
        pass


    # Enter a parse tree produced by Ah210Parser#flow.
    def enterFlow(self, ctx:Ah210Parser.FlowContext):
        pass

    # Exit a parse tree produced by Ah210Parser#flow.
    def exitFlow(self, ctx:Ah210Parser.FlowContext):
        pass


    # Enter a parse tree produced by Ah210Parser#if_statement.
    def enterIf_statement(self, ctx:Ah210Parser.If_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#if_statement.
    def exitIf_statement(self, ctx:Ah210Parser.If_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#while_statement.
    def enterWhile_statement(self, ctx:Ah210Parser.While_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#while_statement.
    def exitWhile_statement(self, ctx:Ah210Parser.While_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#for_statement.
    def enterFor_statement(self, ctx:Ah210Parser.For_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#for_statement.
    def exitFor_statement(self, ctx:Ah210Parser.For_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#each_statement.
    def enterEach_statement(self, ctx:Ah210Parser.Each_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#each_statement.
    def exitEach_statement(self, ctx:Ah210Parser.Each_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#condition.
    def enterCondition(self, ctx:Ah210Parser.ConditionContext):
        pass

    # Exit a parse tree produced by Ah210Parser#condition.
    def exitCondition(self, ctx:Ah210Parser.ConditionContext):
        pass


    # Enter a parse tree produced by Ah210Parser#block.
    def enterBlock(self, ctx:Ah210Parser.BlockContext):
        pass

    # Exit a parse tree produced by Ah210Parser#block.
    def exitBlock(self, ctx:Ah210Parser.BlockContext):
        pass


    # Enter a parse tree produced by Ah210Parser#callback.
    def enterCallback(self, ctx:Ah210Parser.CallbackContext):
        pass

    # Exit a parse tree produced by Ah210Parser#callback.
    def exitCallback(self, ctx:Ah210Parser.CallbackContext):
        pass


    # Enter a parse tree produced by Ah210Parser#callback_block.
    def enterCallback_block(self, ctx:Ah210Parser.Callback_blockContext):
        pass

    # Exit a parse tree produced by Ah210Parser#callback_block.
    def exitCallback_block(self, ctx:Ah210Parser.Callback_blockContext):
        pass


    # Enter a parse tree produced by Ah210Parser#optional_parameters_block.
    def enterOptional_parameters_block(self, ctx:Ah210Parser.Optional_parameters_blockContext):
        pass

    # Exit a parse tree produced by Ah210Parser#optional_parameters_block.
    def exitOptional_parameters_block(self, ctx:Ah210Parser.Optional_parameters_blockContext):
        pass


    # Enter a parse tree produced by Ah210Parser#sig_parameter.
    def enterSig_parameter(self, ctx:Ah210Parser.Sig_parameterContext):
        pass

    # Exit a parse tree produced by Ah210Parser#sig_parameter.
    def exitSig_parameter(self, ctx:Ah210Parser.Sig_parameterContext):
        pass


    # Enter a parse tree produced by Ah210Parser#call_parameter.
    def enterCall_parameter(self, ctx:Ah210Parser.Call_parameterContext):
        pass

    # Exit a parse tree produced by Ah210Parser#call_parameter.
    def exitCall_parameter(self, ctx:Ah210Parser.Call_parameterContext):
        pass


    # Enter a parse tree produced by Ah210Parser#optional_parameter.
    def enterOptional_parameter(self, ctx:Ah210Parser.Optional_parameterContext):
        pass

    # Exit a parse tree produced by Ah210Parser#optional_parameter.
    def exitOptional_parameter(self, ctx:Ah210Parser.Optional_parameterContext):
        pass


    # Enter a parse tree produced by Ah210Parser#commandtax.
    def enterCommandtax(self, ctx:Ah210Parser.CommandtaxContext):
        pass

    # Exit a parse tree produced by Ah210Parser#commandtax.
    def exitCommandtax(self, ctx:Ah210Parser.CommandtaxContext):
        pass


    # Enter a parse tree produced by Ah210Parser#execute.
    def enterExecute(self, ctx:Ah210Parser.ExecuteContext):
        pass

    # Exit a parse tree produced by Ah210Parser#execute.
    def exitExecute(self, ctx:Ah210Parser.ExecuteContext):
        pass


    # Enter a parse tree produced by Ah210Parser#async_execute.
    def enterAsync_execute(self, ctx:Ah210Parser.Async_executeContext):
        pass

    # Exit a parse tree produced by Ah210Parser#async_execute.
    def exitAsync_execute(self, ctx:Ah210Parser.Async_executeContext):
        pass


    # Enter a parse tree produced by Ah210Parser#await.
    def enterAwait(self, ctx:Ah210Parser.AwaitContext):
        pass

    # Exit a parse tree produced by Ah210Parser#await.
    def exitAwait(self, ctx:Ah210Parser.AwaitContext):
        pass


    # Enter a parse tree produced by Ah210Parser#labels.
    def enterLabels(self, ctx:Ah210Parser.LabelsContext):
        pass

    # Exit a parse tree produced by Ah210Parser#labels.
    def exitLabels(self, ctx:Ah210Parser.LabelsContext):
        pass


    # Enter a parse tree produced by Ah210Parser#label_comp.
    def enterLabel_comp(self, ctx:Ah210Parser.Label_compContext):
        pass

    # Exit a parse tree produced by Ah210Parser#label_comp.
    def exitLabel_comp(self, ctx:Ah210Parser.Label_compContext):
        pass


    # Enter a parse tree produced by Ah210Parser#params_statement.
    def enterParams_statement(self, ctx:Ah210Parser.Params_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#params_statement.
    def exitParams_statement(self, ctx:Ah210Parser.Params_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#options_statement.
    def enterOptions_statement(self, ctx:Ah210Parser.Options_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#options_statement.
    def exitOptions_statement(self, ctx:Ah210Parser.Options_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#delete_statement.
    def enterDelete_statement(self, ctx:Ah210Parser.Delete_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#delete_statement.
    def exitDelete_statement(self, ctx:Ah210Parser.Delete_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#error_statement.
    def enterError_statement(self, ctx:Ah210Parser.Error_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#error_statement.
    def exitError_statement(self, ctx:Ah210Parser.Error_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#return_statement.
    def enterReturn_statement(self, ctx:Ah210Parser.Return_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#return_statement.
    def exitReturn_statement(self, ctx:Ah210Parser.Return_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#login_statement.
    def enterLogin_statement(self, ctx:Ah210Parser.Login_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#login_statement.
    def exitLogin_statement(self, ctx:Ah210Parser.Login_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#endpoint_statement.
    def enterEndpoint_statement(self, ctx:Ah210Parser.Endpoint_statementContext):
        pass

    # Exit a parse tree produced by Ah210Parser#endpoint_statement.
    def exitEndpoint_statement(self, ctx:Ah210Parser.Endpoint_statementContext):
        pass


    # Enter a parse tree produced by Ah210Parser#scoping.
    def enterScoping(self, ctx:Ah210Parser.ScopingContext):
        pass

    # Exit a parse tree produced by Ah210Parser#scoping.
    def exitScoping(self, ctx:Ah210Parser.ScopingContext):
        pass


    # Enter a parse tree produced by Ah210Parser#name.
    def enterName(self, ctx:Ah210Parser.NameContext):
        pass

    # Exit a parse tree produced by Ah210Parser#name.
    def exitName(self, ctx:Ah210Parser.NameContext):
        pass


    # Enter a parse tree produced by Ah210Parser#exports.
    def enterExports(self, ctx:Ah210Parser.ExportsContext):
        pass

    # Exit a parse tree produced by Ah210Parser#exports.
    def exitExports(self, ctx:Ah210Parser.ExportsContext):
        pass


    # Enter a parse tree produced by Ah210Parser#imports.
    def enterImports(self, ctx:Ah210Parser.ImportsContext):
        pass

    # Exit a parse tree produced by Ah210Parser#imports.
    def exitImports(self, ctx:Ah210Parser.ImportsContext):
        pass


    # Enter a parse tree produced by Ah210Parser#casting.
    def enterCasting(self, ctx:Ah210Parser.CastingContext):
        pass

    # Exit a parse tree produced by Ah210Parser#casting.
    def exitCasting(self, ctx:Ah210Parser.CastingContext):
        pass


    # Enter a parse tree produced by Ah210Parser#auth.
    def enterAuth(self, ctx:Ah210Parser.AuthContext):
        pass

    # Exit a parse tree produced by Ah210Parser#auth.
    def exitAuth(self, ctx:Ah210Parser.AuthContext):
        pass


    # Enter a parse tree produced by Ah210Parser#url.
    def enterUrl(self, ctx:Ah210Parser.UrlContext):
        pass

    # Exit a parse tree produced by Ah210Parser#url.
    def exitUrl(self, ctx:Ah210Parser.UrlContext):
        pass


    # Enter a parse tree produced by Ah210Parser#log.
    def enterLog(self, ctx:Ah210Parser.LogContext):
        pass

    # Exit a parse tree produced by Ah210Parser#log.
    def exitLog(self, ctx:Ah210Parser.LogContext):
        pass


    # Enter a parse tree produced by Ah210Parser#count.
    def enterCount(self, ctx:Ah210Parser.CountContext):
        pass

    # Exit a parse tree produced by Ah210Parser#count.
    def exitCount(self, ctx:Ah210Parser.CountContext):
        pass


    # Enter a parse tree produced by Ah210Parser#inject.
    def enterInject(self, ctx:Ah210Parser.InjectContext):
        pass

    # Exit a parse tree produced by Ah210Parser#inject.
    def exitInject(self, ctx:Ah210Parser.InjectContext):
        pass


    # Enter a parse tree produced by Ah210Parser#atom.
    def enterAtom(self, ctx:Ah210Parser.AtomContext):
        pass

    # Exit a parse tree produced by Ah210Parser#atom.
    def exitAtom(self, ctx:Ah210Parser.AtomContext):
        pass


    # Enter a parse tree produced by Ah210Parser#obj_dict.
    def enterObj_dict(self, ctx:Ah210Parser.Obj_dictContext):
        pass

    # Exit a parse tree produced by Ah210Parser#obj_dict.
    def exitObj_dict(self, ctx:Ah210Parser.Obj_dictContext):
        pass


    # Enter a parse tree produced by Ah210Parser#obj_list.
    def enterObj_list(self, ctx:Ah210Parser.Obj_listContext):
        pass

    # Exit a parse tree produced by Ah210Parser#obj_list.
    def exitObj_list(self, ctx:Ah210Parser.Obj_listContext):
        pass


    # Enter a parse tree produced by Ah210Parser#string.
    def enterString(self, ctx:Ah210Parser.StringContext):
        pass

    # Exit a parse tree produced by Ah210Parser#string.
    def exitString(self, ctx:Ah210Parser.StringContext):
        pass


    # Enter a parse tree produced by Ah210Parser#number.
    def enterNumber(self, ctx:Ah210Parser.NumberContext):
        pass

    # Exit a parse tree produced by Ah210Parser#number.
    def exitNumber(self, ctx:Ah210Parser.NumberContext):
        pass


    # Enter a parse tree produced by Ah210Parser#boolean.
    def enterBoolean(self, ctx:Ah210Parser.BooleanContext):
        pass

    # Exit a parse tree produced by Ah210Parser#boolean.
    def exitBoolean(self, ctx:Ah210Parser.BooleanContext):
        pass





from apitax.drivers.plugins.BasicGit import BasicGitDriver
from apitax.drivers.plugins.BasicAuth import BasicAuthDriver

from apitax.drivers.plugins.ApitaxTests import ApitaxTestsDriver
from apitax.drivers.plugins.commandtax.ApitaxTestsCommands import ApitaxTestsCommands
from apitax.drivers.plugins.ApitaxInfo import ApitaxInfoDriver
from apitax.drivers.plugins.commandtax.ApitaxInfoCommands import ApitaxInfoCommands

from apitax.drivers.plugins.Openstack import OpenstackDriver

from apitax.drivers.plugins.Github import GithubDriver


from apitax.logs.Log import Log


class Drivers:

    drivers = None

    @staticmethod
    def initialize():
        Drivers.drivers = {
            "BasicGitDriver": BasicGitDriver(),
            "BasicAuthDriver": BasicAuthDriver(),

            "ApitaxTestsDriver": ApitaxTestsDriver(),
            "ApitaxTestsCommands": ApitaxTestsCommands(),
            "ApitaxInfoDriver": ApitaxInfoDriver(),
            "ApitaxInfoCommands": ApitaxInfoCommands(),

            "OpenstackDriver": OpenstackDriver(),

            "GithubDriver": GithubDriver(),
        }

    @staticmethod
    def add(name, driver):
        if(name == 'Default' or name == 'DefaultDriver'):
            Log().error("A driver cannot be named default for driver: " + name)
            return None
        Drivers.drivers[name] = driver

    @staticmethod
    def get(name):
        if (name not in Drivers.drivers):
            Log().error("Driver '" + name + "' does not exist or has not been imported/added.")
            return None
        return Drivers.drivers[name]



import json
import traceback

import connexion
import six

import flask

from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.execute import Execute  # noqa: E501
from apitax.ah.api.models.response import Response  # noqa: E501
from apitax.ah.api.models.user_auth import UserAuth  # noqa: E501
from apitax.ah.api import util

from apitax.ah.State import State
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.ah.Credentials import Credentials
from apitax.ah.Options import Options
from apitax.ah.Connector import Connector
from apitax.ah.api.utilities.Mappers import mapUserAuthToCredentials

from apitax.ah.api.utilities.Roles import hasAccess as roleAccess
from flask import redirect

from flask_jwt_extended import (create_access_token, create_refresh_token, jwt_required, jwt_refresh_token_required,
                                get_jwt_identity, get_raw_jwt, get_jwt_claims)


@jwt_required
def command(execute=None):  # noqa: E501
    """Execute a Command

    Execute a command # noqa: E501

    :param execute: The data needed to execute this command
    :type execute: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        execute = Execute.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    try:
        connector = None

        parameters = {}

        if (execute.command.parameters):
            parameters = execute.command.parameters

        credentials = Credentials()
        options = Options(debug=execute.command.options['debug'], sensitive=execute.command.options['sensitive'])

        if (execute.auth):
            credentials = mapUserAuthToCredentials(execute.auth, credentials)

        if (not execute.auth.api_token):
            options.sensitive = True

        connector = Connector(options=options, credentials=credentials, command=execute.command.command,
                              parameters=parameters)

        commandHandler = connector.execute()

        response = Response(status=commandHandler.getRequest().getResponseStatusCode(),
                            body=json.loads(commandHandler.getRequest().getResponseBody()))

        if (execute.command.options['debug']):
            response.log = connector.logBuffer

        return response
    except:
        State.log.error(traceback.format_exc())
        if ('debug' in execute.command.options and execute.command.options['debug']):
            return ErrorResponse(status=500,
                                 message="Uncaught exception occured during processing. To get a larger stack trace, visit the logs.",
                                 state=traceback.format_exc(3))
        else:
            return ErrorResponse(status=500, message="")


#@jwt_required
def display_dashboard():  # noqa: E501
    """Displays the user dashboard page

    Displays the user dashboard page # noqa: E501


    :rtype: None
    """

 #   if(not hasAccess()):
  #      return redirectUnauthorized()

    return flask.send_from_directory('dashboard/src/pages', 'user_dashboard.html')


@jwt_required
def endpoint_catalog(catalog=None):  # noqa: E501
    """Retrieve the endpoint catalog

    Retrieve the endpoint catalog # noqa: E501

    :param catalog: The data needed to get a catalog
    :type catalog: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        catalog = UserAuth.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    auth = None
    if(catalog):
        auth = catalog
    return Response(status=200, body=driver.getCatalog(auth))


@jwt_required
def get_script(name=None):  # noqa: E501
    """Retrieve the contents of a script

    Retrieve the contents of a script # noqa: E501

    :param name: The script name.
    :type name: str

    :rtype: Response
    """

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    return Response(status=200, body=driver.readScript(name))


@jwt_required
def script_catalog():  # noqa: E501
    """Retrieve the script catalog

    Retrieve the script catalog # noqa: E501


    :rtype: Response
    """
    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    return Response(status=200, body=driver.getScriptsCatalog())


def hasAccess():
    if (roleAccess(get_jwt_identity(), 'developer')):
        return True
    return False


def redirectUnauthorized(page="login", code=303):
    return redirect(State.baseUrl + page, code=code)


from apitax.integrations.Github import Github
from apitax.drivers.Driver import Driver
from apitax.utilities.Files import getAllFiles
from apitax.utilities.Files import getPath
from apitax.utilities.Files import createDir
from pathlib import Path


class GithubDriver(Driver):

    def __init__(self):
        super().__init__()
        self.git = Github(self.driverConfig.get('personal-access-token'), self.driverConfig.get('repo'))

    def isApiAuthenticated(self):
        return False

    def isTokenable(self):
        return False

    def getScriptsCatalog(self):
        files = self.git.getFiles()
        returner = {"scripts": []}
        for file in files:
            path = file.path
            if (path[-3:] == '.ah'):
                returner['scripts'].append({"label": path.split('/')[-1].split('.')[0].title(), "relative-path": path,
                                            "path": self.git.getPath(path)})
        return returner

    def readScript(self, path):
        return self.git.getFileContent(path)

    def renameScript(self, pathOriginal, pathNew):
        return self.git.renameFile(pathOriginal, pathNew)

    # Save does update and create
    def saveScript(self, path, content):
        return self.git.createOrUpdate(path, content)

    def deleteScript(self, path):
        return self.git.deleteFile(path=path)





class LogDriver:
	
    WHITE = '\033[97m'
    CYAN = '\033[96m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
	
	
    def __init__ (self):
        self.logger = None
        self.settings = None
        self.buffer = []
        
    def setSettings(self, settings):
        self.settings = settings
        
    def setLogger(self, logger):
        self.logger = logger
        
    def log(self, line, type='info'):
        pass
        
    def outputLog(self):
        pass
        
    def injectStdColor(self, text):
        text = str(text)
        
        if(not self.settings.get('colorize')):
            return text
        
        if(text[:3] == '>>>'):
            return self.HEADER + text + self.ENDC
        elif(text[:2] == '>>'):
            return self.OKBLUE + text + self.ENDC
        elif(text[:1] == '>'):
            if(text[:17] == '> Now processing:'):
                return self.OKGREEN + text + self.ENDC
            else:
                return self.WARNING + text + self.ENDC
        elif(text[:3] == '###'):
            return self.FAIL + text + self.ENDC
        elif(text.strip()[:1] == '*'):
            return self.CYAN + text + self.ENDC
            
        return self.WHITE + text + self.ENDC
        
    def stripColor(self, text):
        ansi_escape = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')
        return ansi_escape.sub('', text)


import sys
from antlr4 import *
from apitax.grammar.build.Ah210Lexer import Ah210Lexer
from apitax.grammar.build.Ah210Parser import Ah210Parser
from apitax.grammar.AhListener import AhListener
from apitax.grammar.AhVisitor import AhVisitor


def GrammarTest(filepath):
    # input = FileStream(argv[1])
    input = FileStream(filepath)
    lexer = Ah210Lexer(input)
    stream = CommonTokenStream(lexer)
    parser = Ah210Parser(stream)
    tree = parser.prog()
    printer = AhListener()

    visitor = AhVisitor(None, None)
    result = visitor.visit(tree)
    # print(result)

    # walker = ParseTreeWalker()
    # walker.walk(printer, tree)


if __name__ == '__main__':
    GrammarTest(sys.argv)



import configparser
from apitax.utilities.Numbers import isNumber
from apitax.utilities.Booleans import isBoolean, str2bool
from apitax.utilities.Files import getRoot
from apitax.ah.State import State

# Reads the config and allows us to retrieve values from it
class Config:
	
    configs = {}
    
    def __init__(self, path, sectionName):
        self.cp = configparser.RawConfigParser()
        self.cFilePath = path
        self.sectionName = sectionName
        self.cp.read(self.cFilePath)
        self.path = ''

    @staticmethod
    def read(path="", sectionName='Apitax'):
        if(State.paths['config'] != "" and path == ""):
            path = State.paths['config']
        elif(path == ""):
            path = getRoot('/config.txt')
        if(path+sectionName in Config.configs):
            return Config.configs[path+sectionName]
        config = Config(path, sectionName)
        Config.configs[path+sectionName] = config
        return config

    def get(self, param):
        prop: str = self.cp.get(self.sectionName, param)
        if(isBoolean(prop)):
            return str2bool(prop)
        elif(isNumber(prop)):
            return float(prop)
        return prop
        
    def getAsList(self, param):
        items = list(str(self.cp.get(self.sectionName, param)).split(","))
        return [x.strip(' ') for x in items]

    def has(self, param):
        return self.cp.has_option(self.sectionName, param)
        
    def serialize(self, serializable:list):
        returner = dict()
        for property in serializable:
            if(self.has(property)):
                returner.update({property: self.get(property)})
        return returner



from apitax.drivers.Drivers import Drivers
from apitax.logs.Log import Log
from apitax.ah.State import State


class LoadedDrivers:
    drivers = {}

    default = {}

    @staticmethod
    def load(name):
        setDefaultDrivers = False
        if (not LoadedDrivers.default):
            setDefaultDrivers = True

        if (name + 'Driver' not in Drivers.drivers):
            Log().error("Driver '" + name + 'Driver' + "' does not exist or has not been imported/added.")
            return

        LoadedDrivers.drivers[name + 'Driver'] = Drivers.get(name + 'Driver')

        if (setDefaultDrivers):
            LoadedDrivers.default['base'] = Drivers.get(name + 'Driver')

        Log().log("> Driver '" + name + 'Driver' + "' is loaded.")
        Log().log('')

        if (name + 'Commands' in Drivers.drivers):
            LoadedDrivers.drivers[name + 'Commands'] = Drivers.get(name + 'Commands')
            Log().log("> Driver '" + name + 'Commands' + "' is loaded.")
            Log().log('')

            if (setDefaultDrivers):
                LoadedDrivers.default['commands'] = Drivers.get(name + 'Commands')

    @staticmethod
    def getDefaultBaseDriver():
        return LoadedDrivers.default['base']

    @staticmethod
    def getDefaultCommandsDriver():
        if ('commands' in LoadedDrivers.default):
            return LoadedDrivers.default['commands']
        else:
            Log().error('No default command driver is loaded')

    @staticmethod
    def getPrimaryDriver():
        if (not State.config.has("drivers-primary") or State.config.get("drivers-primary") == "default"):
            return LoadedDrivers.getDefaultBaseDriver()
        return LoadedDrivers.getBaseDriver(State.config.get('drivers-primary') + 'Driver')

    @staticmethod
    def getAuthDriver():
        if (not State.config.has("drivers-auth") or State.config.get("drivers-auth") == "default"):
            return LoadedDrivers.getDefaultBaseDriver()
        return LoadedDrivers.getBaseDriver(State.config.get('drivers-auth') + 'Driver')

    @staticmethod
    def getBaseDriver(name):
        if (name not in Drivers.drivers):
            Log().error("Driver '" + name + "' has not been loaded.")
        return LoadedDrivers.drivers[name]

    @staticmethod
    def getCommandsDriver(name):
        if (name + 'Commands' not in Drivers.drivers):
            Log().error("Driver '" + name + 'Commands' + "' has not been loaded.")
        return LoadedDrivers.drivers[name + 'Commands']



# System import
import shlex

# Application import
from apitax.ah.commandtax.commands.Script import Script
from apitax.ah.commandtax.commands.Custom import Custom
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.ah.Credentials import Credentials
from apitax.ah.Options import Options


# Command is used to distribute the workload amoung a heirarchy of possible handlers
# Command is the 'brain' of the application
class Commandtax:
    def __init__(self, header, command, config, options=Options(), parameters={}, auth=Credentials()):

        if (type(command) is not list):
            command = shlex.split(command.strip())
        if (len(command) < 1):
            return
        self.request = None

        if ('--debug' in command):
            options.debug = True
        if ('--sensitive' in command):
            options.sensitive = True
        if ('--driver' in command):
            options.driver = command[command.index('--driver') + 1]

        if (command[0] == 'script'):
            self.request = Script(config, header, auth, parameters, options)
        elif (command[0] == 'custom'):
            self.request = Custom(config, header, auth, parameters, options)
        else:
            if (options.driver):
                customCommands = LoadedDrivers.getCommandsDriver(options.driver)
            else:
                customCommands = LoadedDrivers.getDefaultCommandsDriver()
            self.request = customCommands.setup(config, header, auth, parameters, options)
            self.request = self.request.handle(command)
            return

        self.request.handle(command[1:])

    def getRequest(self):
        return self.request

    def getReturnedData(self):
        returned = None
        if (isinstance(self.getRequest(), Script)):
            returned = self.getRequest().parser.data.getReturn()
        if (returned is not None):
            return returned
        else:
            return self.getRequest().getResponseBody()



# Library import
import click

# Application import
from apitax.ah.commandtax.commands.Custom import Custom
from apitax.ah.HeaderBuilder import HeaderBuilder
from apitax.ah.Options import Options
from apitax.ah.State import State


# Handles creating an authentication request for tokenable drivers
# It does this by creating a command to an authentication endpoint with the applicable data
class AuthRequest(Custom):
    def __init__(self, credentials, driver, options):
        self.driver = driver
        self.log = State.log
        config = State.config

        if (not self.driver.isApiAuthenticated()):
            return

        header = HeaderBuilder()
        header.build(self.driver.getContentTypeJSON())

        if (not self.driver.isCredentialsPosted):
            header.build(self.driver.getPasswordAuthHeader(credentials))

        Custom.__init__(self, config, header, None, None, Options(debug=options.debug, sensitive=True))

        if (self.driver.isCredentialsPosted):
            self.setPostData(self.driver.getPasswordAuthData(credentials))

        if (self.options.debug):
            self.log.log('>> AuthRequest Created')

        self.custom = '--post --url ' + self.driver.getAuthEndpoint() + ' --data-path \'{"user":"' + credentials.username + '"}\''

    def isAuthenticated(self):
        if (not self.driver.isApiAuthenticated()):
            return True
        if(self.driver.isTokenable() and self.getToken()):
            return True
        if(self.getResponseStatusCode() >= 200 and self.getResponseStatusCode() < 300):
            return True
        return False

    def authenticate(self):
        if (not self.driver.isApiAuthenticated()):
            return
        self.handle(self.custom)

    def getToken(self):
        if (not self.driver.isApiAuthenticated()):
            return None
        return self.driver.getToken(self)  # request.headers.get('X-Subject-Token')



import connexion
import six

from apitax.ah.api.models.create import Create  # noqa: E501
from apitax.ah.api.models.delete import Delete  # noqa: E501
from apitax.ah.api.models.error_response import ErrorResponse  # noqa: E501
from apitax.ah.api.models.rename import Rename  # noqa: E501
from apitax.ah.api.models.response import Response  # noqa: E501
from apitax.ah.api.models.save import Save  # noqa: E501
from apitax.ah.api import util

from apitax.ah.State import State
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.ah.api.utilities.Roles import isRole

from apitax.ah.api.utilities.Roles import hasAccess as roleAccess
from flask import redirect

from flask_jwt_extended import (create_access_token, create_refresh_token, jwt_required, jwt_refresh_token_required,
                                get_jwt_identity, get_raw_jwt, get_jwt_claims)

@jwt_required
def create_script(create=None):  # noqa: E501
    """Create a new script

    Create a new script # noqa: E501

    :param Scripts: The data needed to create this script
    :type Scripts: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        create = Create.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    driver.saveScript(create.script.name, create.script.content)
    return Response(status=200, body={'file-name': create.script.name})


@jwt_required
def delete_script(delete=None):  # noqa: E501
    """Delete a script

    Delete a script # noqa: E501

    :param delete: The data needed to delete this script
    :type delete: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        delete = Delete.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    driver.deleteScript(delete.script.name)
    return Response(status=200, body={})


@jwt_required
def display_developer_dashboard():  # noqa: E501
    """Displays the developer dashboard page

    Displays the user dashboard page # noqa: E501


    :rtype: None
    """

    if(not hasAccess()):
        return redirectUnauthorized()

    return 'do some magic!'


@jwt_required
def rename_script(rename=None):  # noqa: E501
    """Rename a script

    Rename a script # noqa: E501

    :param rename: The data needed to save this script
    :type rename: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        rename = Rename.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    if (not driver.renameScript(rename.original.name, rename.new.name)):
        return ErrorResponse(status=500, message='Cannot rename to an existing file.')
    return Response(status=200, body={'file-name': rename.new.name})


@jwt_required
def save_script(save=None):  # noqa: E501
    """Save a script

    Save a script # noqa: E501

    :param Scripts: The data needed to save this script
    :type Scripts: dict | bytes

    :rtype: Response
    """
    if connexion.request.is_json:
        save = Save.from_dict(connexion.request.get_json())  # noqa: E501

    if(not hasAccess()):
        return redirectUnauthorized()

    driver = LoadedDrivers.getDefaultBaseDriver()
    driver.saveScript(save.script.name, save.script.content)
    return Response(status=200, body={'file-name': save.script.name})

def hasAccess():
    if (roleAccess(get_jwt_identity(), 'developer')):
        return True
    return False


def redirectUnauthorized(page="login", code=303):
    return redirect(State.baseUrl + page, code=code)



from apitax.grammar.build.Ah210Parser import Ah210Parser
from apitax.grammar.build.Ah210Visitor import Ah210Visitor as Ah210VisitorOriginal
from apitax.ah.scriptax.ScriptData import ScriptData as DataStore
from apitax.logs.Log import Log
from apitax.utilities.Async import GenericExecution
from apitax.utilities.Json import isJson
from apitax.ah.LoadedDrivers import LoadedDrivers
from apitax.ah.Credentials import Credentials
from apitax.ah.Options import Options

import json
import re
import threading


class AhVisitor(Ah210VisitorOriginal):

    def __init__(self, config, header, auth, parameters={}, options=Options(), file=''):
        self.data = DataStore()
        self.log = Log()
        self.appOptions = options
        self.header = header
        self.config = config
        self.parser = None
        self.options = {}
        self.data.setAuth(auth)
        self.data.storeVar("params.passed", parameters)
        self.state = {'file': file, 'line': 0, 'char': 0}
        self.threads = []

        # Replace the below functionality if possible
        self.regexVar = '{{[ ]{0,}[A-z0-9_$.\-]{1,}[ ]{0,}}}'

    def setState(self, file='', line=-1, char=-1):
        if (file != ''):
            self.state['file'] = file
        if (line != -1):
            self.state['line'] = line
        if (char != -1):
            self.state['char'] = char

    def importCommandRequest(self, commandHandler, export=False):
        from apitax.ah.commandtax.commands.Custom import Custom as CustomCommand

        if (commandHandler.getRequest().getResponseBody().strip() != ''):
            if (not isinstance(commandHandler.getRequest(), CustomCommand)):
                self.data.importScriptsExports(commandHandler.getRequest().parser.data, export=export)
            else:
                self.data.storeRequest(commandHandler.getRequest().getResponseBody(), export=export)

    def executeCommand(self, resolvedCommand, logPrefix=''):
        from apitax.ah.Connector import Connector
        if (self.appOptions.debug):
            self.log.log('> Executing Commandtax: \'' + resolvedCommand['command'] + '\' ' + 'with parameters: ' + str(
                resolvedCommand['parameters']), logPrefix)
            self.log.log('')

        auth = None
        if (resolvedCommand['auth']):
            auth = resolvedCommand['auth']
        else:
            auth = self.data.getAuth()

        connector = Connector(options=Options(debug=self.appOptions.debug, sensitive=self.appOptions.sensitive,
                                              driver=resolvedCommand['driver']),
                              credentials=Credentials(username=auth.username, password=auth.password, token=auth.token),
                              command=resolvedCommand['command'], parameters=resolvedCommand['parameters'], json=True)
        commandHandler = connector.execute()

        if (hasattr(commandHandler.getRequest(), 'parser')):
            if (commandHandler.getRequest().parser.isError()):
                self.error('Subscript contains error: ' + commandHandler.getRequest().parser.isError()['message'],
                           logPrefix)

        if (resolvedCommand['strict'] and commandHandler.getRequest().getResponseStatusCode() >= 300):
            self.error('Request returned non-success status code while in strict mode. Request returned: Status ' +
                       str(
                           commandHandler.getRequest().getResponseStatusCode()) + ' ' + commandHandler.getRequest().getResponseBody(),
                       logPrefix)

        returnResult = commandHandler.getReturnedData()
        if ('callback' in resolvedCommand):
            returnResult = self.executeIsolatedCallback(resolvedCommand['callback'], returnResult, logPrefix)

        return dict({"command": resolvedCommand['command'], "commandHandler": commandHandler, "result": returnResult})

    def getVariable(self, label, isRequest=False, convert=True):
        if (convert):
            label = self.visit(label)
        try:
            if (isRequest):
                return self.data.getRequest(label)
            else:
                return self.data.getVar(label)
        except:

            return None

    def useOptions(self):
        if (self.options['params']):

            if (len(self.options['params']) != len(self.data.getVar('params.passed'))):
                self.error(
                    'Insufficient parameters. Expected: ' + str(self.options['params']) + ' but received: ' + str(
                        self.data.getVar('params.passed')))
            else:
                i = 0
                for param in self.options['params']:
                    self.data.storeVar('params.' + param, self.data.getVar('params.passed.' + param))
                    i += 1

    def inject(self, line):
        matches = re.findall(self.regexVar, line)
        for match in matches:
            label = match[2:-2].strip()
            replacer = self.getVariable(label, convert=False)
            line = line.replace(match, replacer)
        return line

    def executeIsolatedCallback(self, callback, resultScope, logPrefix=''):
        visitor = AhVisitor(self.config, self.header, self.data.getAuth(), options=Options(debug=self.appOptions.debug,
                                                                                           sensitive=self.appOptions.sensitive))
        visitor.setState(file=self.state['file'])
        visitor.log.prefix = logPrefix
        visitor.data.storeVar('result', resultScope)
        for key, value in callback['params'].items():
            visitor.data.storeVar(key, value)
        block = callback['block']
        callbackResult = visitor.visit(block)
        return visitor.data.getVar('result')

    def error(self, message, logprefix=''):
        self.data.error(message, logprefix=logprefix)

    def isError(self):
        return self.data.getFlow('error')

    def isExit(self):
        return self.data.getFlow('exit')

    def isReturn(self):
        return self.data.getFlow('return')

    # Visit a parse tree produced by Ah210Parser#prog.
    def visitProg(self, ctx):
        self.parser = ctx.parser
        temp = self.visitChildren(ctx)

        if (self.isError()):
            error = self.isError()
            self.log.error(error['message'] + ' in ' + self.state['file'] + ' @' + str(self.state['line']) + ':' + str(
                self.state['char']), prefix=error['logprefix'])
            if (self.appOptions.debug):
                self.log.log('')
                self.log.log('')

        return self

    # Visit a parse tree produced by Ah210Parser#statements.
    def visitStatements(self, ctx):
        if (self.isReturn() or self.isError()):
            return
        temp = self.visitChildren(ctx)
        return temp

    # Visit a parse tree produced by Ah210Parser#statement.
    def visitStatement(self, ctx):
        if (self.isReturn()):
            return

        if (self.isError()):
            return

        debugTemp = self.appOptions.debug
        sensitiveTemp = self.appOptions.sensitive

        line = ctx.getText().strip()
        if (line != "" and self.appOptions.debug):
            if (ctx.NOT()):
                self.log.log(
                    '> Now processing: (This lines contents has been hidden via the \'!\' operator. This is usually done to hide sensitive information)')
                self.log.log('')
                self.log.log('> Treating the rest of this statement as sensitive and disabling debug')
                self.log.log('')

                self.appOptions.debug = False
                self.appOptions.sensitive = True
            else:
                self.log.log('> Now processing: ' + line)
                self.log.log('')

        temp = self.visitChildren(ctx)

        if (line != "" and self.appOptions.debug):
            self.log.log('')

        self.setState(line=ctx.start.line)  # TODO: Try to add character here as well

        if (ctx.NOT() and line != "" and debugTemp):
            self.log.log('> Setting debug and sensitive back to their original values')
            self.log.log('')
            self.log.log('')
            self.appOptions.debug = debugTemp
            self.appOptions.sensitive = sensitiveTemp

        return temp

    # Visit a parse tree produced by Ah210Parser#terminated.
    def visitTerminated(self, ctx: Ah210Parser.TerminatedContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#non_terminated.
    def visitNon_terminated(self, ctx: Ah210Parser.Non_terminatedContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#executers.
    def visitExecuters(self, ctx: Ah210Parser.ExecutersContext):
        if (ctx.execute()):
            return self.visit(ctx.execute())
        if (ctx.async_execute()):
            thread = self.visit(ctx.async_execute())
            thread.start()
            return thread

    # Visit a parse tree produced by Ah210Parser#expr.
    def visitExpr(self, ctx):  # Get number of terms and loop this code for #terms - 1
        if (ctx.execute()):
            return self.visit(ctx.execute())['result']

        if (ctx.async_execute()):
            return self.visit(ctx.async_execute())

        if (ctx.atom()):
            return self.visit(ctx.atom())

        if (ctx.casting()):
            return self.visit(ctx.casting())

        if (ctx.login_statement()):
            return self.visit(ctx.login_statement())

        if (ctx.endpoint_statement()):
            return self.visit(ctx.endpoint_statement())

        if (ctx.count()):
            return self.visit(ctx.count())

        if (ctx.labels()):
            return self.getVariable(ctx.labels(), isRequest=ctx.REQUEST())

        if (ctx.inject()):
            return self.visit(ctx.inject())

        if (ctx.MINUS() and not ctx.expr(1)):
            return self.visit(ctx.expr(0)) * -1

        if (ctx.NOT()):
            return not self.visit(ctx.expr(0))

        if (ctx.AND()):
            return self.visit(ctx.expr(0)) and self.visit(ctx.expr(1))

        if (ctx.OR()):
            return self.visit(ctx.expr(0)) or self.visit(ctx.expr(1))

        if (ctx.EQ()):
            return self.visit(ctx.expr(0)) == self.visit(ctx.expr(1))

        if (ctx.NEQ()):
            return self.visit(ctx.expr(0)) != self.visit(ctx.expr(1))

        if (ctx.GE()):
            return self.visit(ctx.expr(0)) >= self.visit(ctx.expr(1))

        if (ctx.LE()):
            return self.visit(ctx.expr(0)) <= self.visit(ctx.expr(1))

        if (ctx.GT()):
            return self.visit(ctx.expr(0)) > self.visit(ctx.expr(1))

        if (ctx.LT()):
            return self.visit(ctx.expr(0)) < self.visit(ctx.expr(1))

        if (ctx.LPAREN()):
            return self.visit(ctx.expr(0))

        if (ctx.POW()):
            return self.visit(ctx.expr(0)) ** self.visit(ctx.expr(1))

        if (ctx.PLUS()):
            left = self.visit(ctx.expr(0))
            right = self.visit(ctx.expr(1))
            if ((isinstance(left, str) and not isinstance(right, str)) or (
                    isinstance(right, str) and not isinstance(left, str))):
                left = str(left)
                right = str(right)
                if (self.appOptions.debug):
                    self.log.log('> Implicit cast to string: \'' + left + '\' + \'' + right + '\'')
                    self.log.log('')
            return left + right

        if (ctx.MINUS()):
            return self.visit(ctx.expr(0)) - self.visit(ctx.expr(1))

        if (ctx.MUL()):
            return self.visit(ctx.expr(0)) * self.visit(ctx.expr(1))

        if (ctx.DIV()):
            return self.visit(ctx.expr(0)) / self.visit(ctx.expr(1))

        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#set_var.
    def visitAssignment(self, ctx: Ah210Parser.AssignmentContext):
        label = self.visit(ctx.labels())
        value = None

        if (ctx.expr()):
            value = self.visit(ctx.expr())

        if (not ctx.EQUAL()):
            var = self.data.getVar(label)
            if (ctx.D_PLUS()):
                value = var + 1
            elif (ctx.D_MINUS()):
                value = var - 1
            else:
                if (ctx.PE()):
                    value += var
                elif (ctx.ME()):
                    value = var - value
                elif (ctx.MUE()):
                    value *= var
                elif (ctx.DE()):
                    value = var / value

        if (ctx.SOPEN()):
            var = self.data.getVar(label)
            if (not isinstance(var, list)):
                self.error("Appending to a list requires the variable being a list")
                return
            var.append(value)
            tval = value
            value = var
            self.data.storeVar(label, value)
            if (isinstance(tval, threading.Thread)):
                tval.label = label + "." + str(len(value) - 1)
                tval.start()
        else:
            self.data.storeVar(label, value)
            if (isinstance(value, threading.Thread)):
                value.label = label
                value.start()

        if (self.appOptions.debug):
            self.log.log('> Assigning Variable: ' + label + ' = ' + str(self.data.getVar(label)))
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#flow.
    def visitFlow(self, ctx: Ah210Parser.FlowContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#if_statement.
    def visitIf_statement(self, ctx: Ah210Parser.If_statementContext):

        i = 0

        while (True):
            if (ctx.condition(i) is None):
                if (ctx.ELSE()):
                    return self.visit(ctx.block(i))
                else:
                    return None
            condition = self.visit(ctx.condition(i))
            if (condition):
                return self.visit(ctx.block(i))
            else:
                i += 1

    # Visit a parse tree produced by Ah210Parser#while_statement.
    def visitWhile_statement(self, ctx: Ah210Parser.While_statementContext):
        while (self.visit(ctx.condition())):
            self.visit(ctx.block())

    # Visit a parse tree produced by Ah210Parser#for_statement.
    def visitFor_statement(self, ctx: Ah210Parser.For_statementContext):
        clause = self.visit(ctx.expr())
        label = self.visit(ctx.labels())

        if (isinstance(clause, str) and isJson(clause)):
            clause = json.loads(clause)

        if (isinstance(clause, list)):
            if (self.appOptions.debug):
                self.log.log('> Looping through list with var ' + label)
                self.log.log('')
            for item in clause:
                if (self.appOptions.debug):
                    self.log.log('>> Assigning ' + label + ' = ' + str(item))
                    self.log.log('')
                    self.log.log('')
                self.data.storeVar(label, item)
                self.visit(ctx.block())
            self.data.deleteVar(label)

        elif (isinstance(clause, float) or isinstance(clause, int)):
            if (self.appOptions.debug):
                self.log.log('> Looping through range with var ' + label)
                self.log.log('')
            for i in range(0, int(clause)):
                if (self.appOptions.debug):
                    self.log.log('>> Assigning ' + label + ' = ' + str(i))
                    self.log.log('')
                    self.log.log('')
                self.data.storeVar(label, i)
                self.visit(ctx.block())
            self.data.deleteVar(label)

        else:
            if (self.appOptions.debug):
                self.error('Invalid Loop Type: ' + str(type(clause)))
                self.log.log('')

    # Visit a parse tree produced by Ah210Parser#each_statement.
    def visitEach_statement(self, ctx: Ah210Parser.Each_statementContext):
        clause = self.visit(ctx.expr())

        if (isinstance(clause, str) and isJson(clause)):
            clause = json.loads(clause)

        if (isinstance(clause, list)):
            if (self.appOptions.debug):
                self.log.log('> Looping through result')
                self.log.log('')
            callback = self.visit(ctx.callback())
            for item in clause:
                if (self.appOptions.debug):
                    self.log.log('>> Assigning result = ' + str(item))
                    self.log.log('')
                    self.log.log('')
                self.executeIsolatedCallback(callback, item)
        else:
            self.error("An each loop must be passed a list")

    # Visit a parse tree produced by Ah210Parser#condition.
    def visitCondition(self, ctx: Ah210Parser.ConditionContext):
        condition = self.visit(ctx.expr())

        if (self.appOptions.debug):
            self.log.log('>> Evaluated Flow Condition as: ' + str(condition))
            self.log.log('')

        return condition

    # Visit a parse tree produced by Ah210Parser#block.
    def visitBlock(self, ctx: Ah210Parser.BlockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#callback.
    def visitCallback(self, ctx: Ah210Parser.CallbackContext):
        parameters = {}
        if (ctx.optional_parameters_block()):
            parameters = self.visit(ctx.optional_parameters_block())
        return {"params": parameters, "block": ctx.callback_block()}

    # Visit a parse tree produced by Ah210Parser#callback_block.
    def visitCallback_block(self, ctx: Ah210Parser.Callback_blockContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#optional_parameters_block.
    def visitOptional_parameters_block(self, ctx: Ah210Parser.Optional_parameters_blockContext):
        i = 0
        parameters = {}
        while (ctx.optional_parameter(i)):
            opParam = self.visit(ctx.optional_parameter(i))
            parameters[opParam['label']] = opParam['value']
            i += 1
        return parameters

    # Visit a parse tree produced by Ah210Parser#sig_parameter.
    def visitSig_parameter(self, ctx: Ah210Parser.Sig_parameterContext):
        if (ctx.labels()):
            return {"label": self.visit(ctx.labels())}
        else:
            return self.visit(ctx.optional_parameter())

    # Visit a parse tree produced by Ah210Parser#call_parameter.
    def visitCall_parameter(self, ctx: Ah210Parser.Call_parameterContext):
        if (ctx.expr()):
            return {"value": self.visit(ctx.expr())}
        else:
            return self.visit(ctx.optional_parameter())

    # Visit a parse tree produced by Ah210Parser#optional_parameter.
    def visitOptional_parameter(self, ctx: Ah210Parser.Optional_parameterContext):
        return {"label": self.visit(ctx.labels()), "value": self.visit(ctx.expr())}

    # Visit a parse tree produced by Ah210Parser#commandtax.
    def visitCommandtax(self, ctx: Ah210Parser.CommandtaxContext):
        from apitax.ah.commandtax.commands.Script import Script as ScriptCommand
        firstArg = self.visit(ctx.expr())
        command = ""
        strict = True
        auth = None
        driver = None

        if (not ctx.SCRIPT() and not ctx.COMMANDTAX()):
            command += "custom"
            if (ctx.GET()):
                command += " --get"
            if (ctx.POST()):
                command += " --post"
            if (ctx.PUT()):
                command += " --put"
            if (ctx.PATCH()):
                command += " --patch"
            if (ctx.DELETE()):
                command += " --delete"
            command += " --url " + self.data.getUrl("current") + firstArg

        elif (ctx.SCRIPT()):
            command += "script " + firstArg

        elif (ctx.COMMANDTAX()):
            command = firstArg

        if (ctx.obj_dict()):
            dataArg = self.visit(ctx.obj_dict())
            if ('post' in dataArg):
                command += " --data-post '" + json.dumps(dataArg['post']) + "'"
            if ('query' in dataArg):
                command += " --data-query '" + json.dumps(dataArg['query']) + "'"
            if ('path' in dataArg):
                command += " --data-path '" + json.dumps(dataArg['path']) + "'"
            if ('header' in dataArg):
                command += " --data-header '" + json.dumps(dataArg['header']) + "'"
            if ('driver' in dataArg):
                command += " --driver " + dataArg['driver']
                driver = dataArg['driver']
            if ('strict' in dataArg):
                strict = bool(dataArg['strict'])
            if ('auth' in dataArg):
                auth = dataArg['auth']

        i = 0
        parameters = {}
        while (ctx.optional_parameter(i)):
            opParam = self.visit(ctx.optional_parameter(i))
            parameters[opParam['label']] = opParam['value']
            i += 1

        return {'command': command, 'parameters': parameters, 'strict': strict, 'auth': auth, 'driver': driver}

    # Visit a parse tree produced by Ah210Parser#execute.
    def visitExecute(self, ctx):
        resolvedCommand = self.visit(ctx.commandtax())
        if (ctx.callback()):
            resolvedCommand['callback'] = self.visit(ctx.callback())
        return self.executeCommand(resolvedCommand)

    # Visit a parse tree produced by Ah210Parser#async_execute.
    def visitAsync_execute(self, ctx: Ah210Parser.Async_executeContext):
        resolvedCommand = self.visit(ctx.commandtax())
        if (ctx.callback()):
            resolvedCommand['callback'] = self.visit(ctx.callback())
        thread = GenericExecution(self, "Async execution and callback", resolvedCommand, log=self.log,
                                  debug=self.appOptions.debug, sensitive=self.appOptions.sensitive)
        self.threads.append(thread)
        return thread

    # Visit a parse tree produced by Ah210Parser#await.
    def visitAwait(self, ctx: Ah210Parser.AwaitContext):
        if (not ctx.labels()):
            for thread in self.threads:
                thread.join()
            return
        threads = self.getVariable(ctx.labels())
        if (isinstance(threads, list)):
            for thread in threads:
                if (isinstance(thread, threading.Thread)):
                    thread.join()
        elif (isinstance(threads, threading.Thread)):
            threads.join()

    # Visit a parse tree produced by Ah210Parser#labels.
    def visitLabels(self, ctx: Ah210Parser.LabelsContext):

        label = [self.visit(ctx.label_comp(0))]
        i = 0
        while (ctx.DOT(i)):
            label.append(str(self.visit(ctx.label_comp(i + 1))))
            i += 1

        label = '.'.join(label)

        return label.replace('$', '')

    # Visit a parse tree produced by Ah210Parser#label_comp.
    def visitLabel_comp(self, ctx: Ah210Parser.Label_compContext):
        if (ctx.LABEL()):
            return ctx.LABEL().getText()
        else:
            return self.visit(ctx.inject())

    # Visit a parse tree produced by Ah210Parser#params_statement.
    def visitParams_statement(self, ctx: Ah210Parser.Params_statementContext):
        i = 0
        while (ctx.sig_parameter(i)):
            sigItem = self.visit(ctx.sig_parameter(i))

            if (self.data.isVarExist('params.passed.' + sigItem['label'])):
                self.data.storeVar('params.' + sigItem['label'], self.data.getVar('params.passed.' + sigItem['label']))
            elif ('value' in sigItem):
                self.data.storeVar('params.' + sigItem['label'], sigItem['value'])
            else:
                self.error(
                    'Insufficient parameters. Expected Parameter: \'' + str(sigItem['label']) + '\'')

            i += 1

    # Visit a parse tree produced by Ah210Parser#options_statement.
    def visitOptions_statement(self, ctx: Ah210Parser.Options_statementContext):
        self.options = self.visit(ctx.expr())
        self.useOptions()

    # Visit a parse tree produced by Ah210Parser#delete_statement.
    def visitDelete_statement(self, ctx: Ah210Parser.Delete_statementContext):
        label = self.visit(ctx.labels())
        self.data.deleteVar(label)
        if (self.appOptions.debug):
            self.log.log('> Deleteing variable: ' + label)
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#error_statement.
    def visitError_statement(self, ctx: Ah210Parser.Error_statementContext):
        message = "No error message was specified"
        if (ctx.expr()):
            message = self.visit(ctx.expr())
        self.error(message)

    # Visit a parse tree produced by Ah210Parser#return_statement.
    def visitReturn_statement(self, ctx: Ah210Parser.Return_statementContext):
        exportation = ""
        if (ctx.expr()):
            exportation = self.visit(ctx.expr())
            self.data.setReturn(exportation)
        else:
            self.data.setReturn({})
        self.data.setFlow('return', True)
        if (self.appOptions.debug):
            if (exportation != ""):
                self.log.log('> Returning with value: ' + str(exportation))
            else:
                self.log.log('> Returning ')
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#login_statement.
    def visitLogin_statement(self, ctx: Ah210Parser.Login_statementContext):
        from apitax.ah.Connector import Connector

        parameters = self.visit(ctx.optional_parameters_block())

        driver = self.appOptions.driver
        extra = {}

        if ('driver' in parameters):
            driver = parameters['driver']

        if ('extra' in parameters):
            extra = parameters['extra']

        if ('username' in parameters and 'password' in parameters):
            if (self.appOptions.debug):
                self.log.log("> Logging into API with username and password.")
                self.log.log("")
            connector = Connector(options=Options(debug=self.appOptions.debug, sensitive=True, driver=driver),
                                  credentials=Credentials(username=parameters['username'],
                                                          password=parameters['password'], extra=extra))
            return connector.getCredentials()
        elif ('token' in parameters):
            return Credentials(token=parameters['token'])
        else:
            self.error('Must pass a `username` and `password` or a `token` for authentication')
            return None

    # Visit a parse tree produced by Ah210Parser#endpoint_statement.
    def visitEndpoint_statement(self, ctx: Ah210Parser.Endpoint_statementContext):
        name = self.visit(ctx.expr())
        driver = None
        try:
            name.find('@')
            name = name.split('@')
            driver = LoadedDrivers.getBaseDriver(name[1])
            name = name[0]
        except:
            if (self.appOptions.driver):
                driver = LoadedDrivers.getBaseDriver(self.appOptions.driver)
            else:
                driver = LoadedDrivers.getDefaultBaseDriver()
        endpoints = driver.getCatalog(self.data.getAuth())['endpoints']
        if (name in endpoints):
            return endpoints[name]['value']
        else:
            self.error("The endpoint requested does not exist")
            return None

    # Visit a parse tree produced by Ah210Parser#scoping.
    def visitScoping(self, ctx):
        # print('scoping')
        return self.visitChildren(ctx)

    # Visit a parse tree produced by Ah210Parser#name.
    def visitName(self, ctx):
        if ctx.expr():
            self.data.name = self.visit(ctx.expr())

        if (self.appOptions.debug):
            self.log.log('> Setting Script Name: ' + self.data.name)
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#exports.
    def visitExports(self, ctx):

        exportation = ""

        if (ctx.labels()):
            exportation = self.visit(ctx.labels())
            self.data.exportVar(exportation)

        if (ctx.execute()):
            commandHandler = self.visit(ctx.execute())
            self.importCommandRequest(commandHandler['commandHandler'], export=True)
            exportation = commandHandler['command']

        if (self.appOptions.debug):
            self.log.log('> Exporting: ' + exportation)
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#imports.
    def visitImports(self, ctx):

        commandHandler = self.visit(ctx.execute())['commandHandler']
        self.importCommandRequest(commandHandler)

        if (self.appOptions.debug):
            self.log.log('> Importing: ' + resolvedCommand['command'])
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#casting.
    def visitCasting(self, ctx: Ah210Parser.CastingContext):
        value = self.visit(ctx.expr())
        if (ctx.TYPE_INT()):
            returner = int(value)
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to int: ' + json.dumps(returner))
                self.log.log('')
            return returner
        if (ctx.TYPE_DEC()):
            returner = float(value)
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to number: ' + json.dumps(returner))
                self.log.log('')
            return returner
        if (ctx.TYPE_BOOL()):
            returner = bool(value)
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to boolean: ' + json.dumps(returner))
                self.log.log('')
            return returner
        if (ctx.TYPE_STR()):
            returner = str(value)
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to string: ' + json.dumps(returner))
                self.log.log('')
            return returner
        if (ctx.TYPE_LIST()):
            returner = list(str(value).split(","))
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to list: ' + json.dumps(returner))
                self.log.log('')
            return returner
        if (ctx.TYPE_DICT()):
            returner = None
            if (isinstance(value, dict)):
                returner = value
            elif (isinstance(value, list)):
                count = 0
                newdict = {}
                for i in value:
                    newdict.update({str(count): i})
                    count += 1
                returner = dict(newdict)
            elif (isinstance(value, str) and isJson(value)):
                returner = dict(json.loads(str(value)))
            else:
                returner = dict({"default": value})
            if (self.appOptions.debug):
                self.log.log('> Explicitly Casting \'' + str(value) + '\' to dictionary: ' + json.dumps(returner))
                self.log.log('')

        return returner

    # Visit a parse tree produced by Ah210Parser#auth.
    def visitAuth(self, ctx: Ah210Parser.AuthContext):
        auth = self.visit(ctx.expr())
        self.data.setAuth(auth)
        if (self.appOptions.debug):
            self.log.log("> Setting active auth credentials to user: " + auth.username)
            self.log.log("")

    # Visit a parse tree produced by Ah210Parser#url.
    def visitUrl(self, ctx: Ah210Parser.UrlContext):
        url = self.visit(ctx.expr())
        self.data.storeUrl("current", url)

        if (self.appOptions.debug):
            self.log.log('> Setting URL: ' + url)
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#log.
    def visitLog(self, ctx: Ah210Parser.LogContext):
        if (ctx.expr()):
            self.log.log('> Logging: ' + json.dumps(self.visit(ctx.expr())))
            self.log.log('')

    # Visit a parse tree produced by Ah210Parser#count.
    def visitCount(self, ctx: Ah210Parser.CountContext):
        return len(self.visit(ctx.expr()))

    # Visit a parse tree produced by Ah210Parser#inject.
    def visitInject(self, ctx: Ah210Parser.InjectContext):
        returner = self.visit(ctx.expr())
        if (self.appOptions.debug):
            self.log.log('> Injecting into: ' + ctx.getText() + ' with the value ' + str(returner))
            self.log.log('')
        return returner

    # Visit a parse tree produced by Ah210Parser#atom.
    def visitAtom(self, ctx: Ah210Parser.AtomContext):
        if (ctx.obj_dict()):
            return self.visit(ctx.obj_dict())

        if (ctx.obj_list()):
            return self.visit(ctx.obj_list())

        if (ctx.string()):
            return self.visit(ctx.string())

        if (ctx.number()):
            return self.visit(ctx.number())

        if (ctx.boolean()):
            return self.visit(ctx.boolean())

    # Visit a parse tree produced by Ah210Parser#obj_dict.
    def visitObj_dict(self, ctx: Ah210Parser.Obj_dictContext):
        dictionary = {}
        i = 0
        if (ctx.COLON(0)):
            dictionary[self.visit(ctx.expr(0))] = self.visit(ctx.expr(1))
        while (ctx.COMMA(i) and ctx.expr((i + 1) * 2)):
            base = (i + 1) * 2
            dictionary[self.visit(ctx.expr(base))] = self.visit(ctx.expr(base + 1))
            i += 1
        return dictionary

    # Visit a parse tree produced by Ah210Parser#obj_list.
    def visitObj_list(self, ctx: Ah210Parser.Obj_listContext):
        parameters = []
        i = 0
        if (ctx.expr(0)):
            parameters.append(self.visit(ctx.expr(0)))
        while (ctx.COMMA(i) and ctx.expr(i + 1)):
            parameters.append(self.visit(ctx.expr(i + 1)))
            i += 1
        return parameters

    # Visit a parse tree produced by Ah210Parser#string.
    def visitString(self, ctx: Ah210Parser.StringContext):
        line = ctx.STRING().getText()[1:-1]
        line = line.replace('\\"', '"')
        line = line.replace('\\\'', '\'')
        matches = re.findall(self.regexVar, line)
        for match in matches:
            label = match[2:-2].strip()
            label = label.replace('$', '')
            replacer = str(self.getVariable(label, convert=False))
            line = line.replace(match, replacer)
            if (self.appOptions.debug):
                self.log.log('> Injecting Variable into String \'' + label + '\': ' + line)
                self.log.log('')
        return line

    # Visit a parse tree produced by Ah210Parser#number.
    def visitNumber(self, ctx: Ah210Parser.NumberContext):
        if (ctx.INT()):
            return int(ctx.INT().getText())
        else:
            return float(ctx.FLOAT().getText())

    # Visit a parse tree produced by Ah210Parser#boolean.
    def visitBoolean(self, ctx: Ah210Parser.BooleanContext):
        if (ctx.TRUE()):
            return True
        if (ctx.FALSE()):
            return False
